import numpy as np
import pandas as pd

from featuretools.primitives import (
    ExponentialWeightedAverage,
    ExponentialWeightedSTD,
    ExponentialWeightedVariance,
)


def test_regular_com_avg():
    primitive_instance = ExponentialWeightedAverage(com=0.5)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series([1.0, 1.75, 5.384615384615384, 5.125])
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_span_avg():
    primitive_instance = ExponentialWeightedAverage(span=1.5)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series([1.0, 1.8333333333333335, 6.0, 5.198717948717948])
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_halflife_avg():
    primitive_instance = ExponentialWeightedAverage(halflife=2.7)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [1.0, 1.563830114594977, 3.8556233149044865, 4.2592901785684205],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_alpha_avg():
    primitive_instance = ExponentialWeightedAverage(alpha=0.8)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series([1.0, 1.8333333333333335, 6.0, 5.198717948717948])
    pd.testing.assert_series_equal(answer, correct_answer)


def test_na_avg():
    primitive_instance = ExponentialWeightedAverage(com=0.5)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, np.nan, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [1.0, 1.75, 5.384615384615384, 5.384615384615384, 5.053191489361702],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_ignorena_true_avg():
    primitive_instance = ExponentialWeightedAverage(com=0.5, ignore_na=True)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, np.nan, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [1.0, 1.75, 5.384615384615384, 5.384615384615384, 5.125],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_com_std():
    primitive_instance = ExponentialWeightedSTD(com=0.5)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [np.nan, 0.7071067811865475, 3.584153156068229, 2.0048019276803304],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_span_std():
    primitive_instance = ExponentialWeightedSTD(span=1.5)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [np.nan, 0.7071067811865476, 3.6055512754639887, 1.7311551816712718],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_halflife_std():
    primitive_instance = ExponentialWeightedSTD(halflife=2.7)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [np.nan, 0.7071067811865475, 3.3565236098585416, 2.631776826295855],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_alpha_std():
    primitive_instance = ExponentialWeightedSTD(alpha=0.8)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [np.nan, 0.7071067811865476, 3.6055512754639887, 1.7311551816712718],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_na_std():
    primitive_instance = ExponentialWeightedSTD(com=0.5)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, np.nan, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [
            np.nan,
            0.7071067811865475,
            3.584153156068229,
            3.5841531560682287,
            1.8408520483016189,
        ],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_ignorena_true_std():
    primitive_instance = ExponentialWeightedSTD(com=0.5, ignore_na=True)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, np.nan, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [
            np.nan,
            0.7071067811865475,
            3.584153156068229,
            3.584153156068229,
            2.0048019276803304,
        ],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_com_var():
    primitive_instance = ExponentialWeightedVariance(com=0.5)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [np.nan, 0.49999999999999983, 12.846153846153847, 4.019230769230769],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_span_var():
    primitive_instance = ExponentialWeightedVariance(span=1.5)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series([np.nan, 0.5, 12.999999999999996, 2.996898263027294])
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_halflife_var():
    primitive_instance = ExponentialWeightedVariance(halflife=2.7)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [np.nan, 0.49999999999999994, 11.266250743537816, 6.926249263427883],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_regular_alpha_var():
    primitive_instance = ExponentialWeightedVariance(alpha=0.8)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series([np.nan, 0.5, 12.999999999999996, 2.996898263027294])
    pd.testing.assert_series_equal(answer, correct_answer)


def test_na_var():
    primitive_instance = ExponentialWeightedVariance(com=0.5)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, np.nan, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [
            np.nan,
            0.49999999999999983,
            12.846153846153847,
            12.846153846153843,
            3.3887362637362655,
        ],
    )
    pd.testing.assert_series_equal(answer, correct_answer)


def test_ignorena_true_var():
    primitive_instance = ExponentialWeightedVariance(com=0.5, ignore_na=True)
    primitive_func = primitive_instance.get_function()
    array = pd.Series([1, 2, 7, np.nan, 5])
    answer = pd.Series(primitive_func(array))
    correct_answer = pd.Series(
        [
            np.nan,
            0.49999999999999983,
            12.846153846153847,
            12.846153846153847,
            4.019230769230769,
        ],
    )
    pd.testing.assert_series_equal(answer, correct_answer)



import pandas as pd
import pytest

from featuretools import EntitySet
from featuretools.tests.testing_utils import backward_path, forward_path


def test_cannot_re_add_relationships_that_already_exists(es):
    before_len = len(es.relationships)
    es.add_relationship(relationship=es.relationships[0])
    after_len = len(es.relationships)
    assert before_len == after_len


def test_add_relationships_convert_type(es):
    for r in es.relationships:
        assert r.parent_dataframe.ww.index == r._parent_column_name
        assert "foreign_key" in r.child_column.ww.semantic_tags
        assert r.child_column.ww.logical_type == r.parent_column.ww.logical_type


def test_get_forward_dataframes(es):
    dataframes = es.get_forward_dataframes("log")
    path_to_sessions = forward_path(es, ["log", "sessions"])
    path_to_products = forward_path(es, ["log", "products"])
    assert list(dataframes) == [
        ("sessions", path_to_sessions),
        ("products", path_to_products),
    ]


def test_get_backward_dataframes(es):
    dataframes = es.get_backward_dataframes("customers")
    path_to_sessions = backward_path(es, ["customers", "sessions"])
    assert list(dataframes) == [("sessions", path_to_sessions)]


def test_get_forward_dataframes_deep(es):
    dataframes = es.get_forward_dataframes("log", deep=True)
    path_to_sessions = forward_path(es, ["log", "sessions"])
    path_to_products = forward_path(es, ["log", "products"])
    path_to_customers = forward_path(es, ["log", "sessions", "customers"])
    path_to_regions = forward_path(es, ["log", "sessions", "customers", "régions"])
    path_to_cohorts = forward_path(es, ["log", "sessions", "customers", "cohorts"])
    assert list(dataframes) == [
        ("sessions", path_to_sessions),
        ("customers", path_to_customers),
        ("cohorts", path_to_cohorts),
        ("régions", path_to_regions),
        ("products", path_to_products),
    ]


def test_get_backward_dataframes_deep(es):
    dataframes = es.get_backward_dataframes("customers", deep=True)
    path_to_log = backward_path(es, ["customers", "sessions", "log"])
    path_to_sessions = backward_path(es, ["customers", "sessions"])
    assert list(dataframes) == [("sessions", path_to_sessions), ("log", path_to_log)]


def test_get_forward_relationships(es):
    relationships = es.get_forward_relationships("log")
    assert len(relationships) == 2
    assert relationships[0]._parent_dataframe_name == "sessions"
    assert relationships[0]._child_dataframe_name == "log"
    assert relationships[1]._parent_dataframe_name == "products"
    assert relationships[1]._child_dataframe_name == "log"

    relationships = es.get_forward_relationships("sessions")
    assert len(relationships) == 1
    assert relationships[0]._parent_dataframe_name == "customers"
    assert relationships[0]._child_dataframe_name == "sessions"


def test_get_backward_relationships(es):
    relationships = es.get_backward_relationships("sessions")
    assert len(relationships) == 1
    assert relationships[0]._parent_dataframe_name == "sessions"
    assert relationships[0]._child_dataframe_name == "log"

    relationships = es.get_backward_relationships("customers")
    assert len(relationships) == 1
    assert relationships[0]._parent_dataframe_name == "customers"
    assert relationships[0]._child_dataframe_name == "sessions"


def test_find_forward_paths(es):
    paths = list(es.find_forward_paths("log", "customers"))
    assert len(paths) == 1

    path = paths[0]

    assert len(path) == 2
    assert path[0]._child_dataframe_name == "log"
    assert path[0]._parent_dataframe_name == "sessions"
    assert path[1]._child_dataframe_name == "sessions"
    assert path[1]._parent_dataframe_name == "customers"


def test_find_forward_paths_multiple_paths(diamond_es):
    paths = list(diamond_es.find_forward_paths("transactions", "regions"))
    assert len(paths) == 2

    path1, path2 = paths

    r1, r2 = path1
    assert r1._child_dataframe_name == "transactions"
    assert r1._parent_dataframe_name == "stores"
    assert r2._child_dataframe_name == "stores"
    assert r2._parent_dataframe_name == "regions"

    r1, r2 = path2
    assert r1._child_dataframe_name == "transactions"
    assert r1._parent_dataframe_name == "customers"
    assert r2._child_dataframe_name == "customers"
    assert r2._parent_dataframe_name == "regions"


def test_find_forward_paths_multiple_relationships(games_es):
    paths = list(games_es.find_forward_paths("games", "teams"))
    assert len(paths) == 2

    path1, path2 = paths
    assert len(path1) == 1
    assert len(path2) == 1
    r1 = path1[0]
    r2 = path2[0]

    assert r1._child_dataframe_name == "games"
    assert r2._child_dataframe_name == "games"
    assert r1._parent_dataframe_name == "teams"
    assert r2._parent_dataframe_name == "teams"

    assert r1._child_column_name == "home_team_id"
    assert r2._child_column_name == "away_team_id"
    assert r1._parent_column_name == "id"
    assert r2._parent_column_name == "id"


@pytest.fixture
def pd_employee_df():
    return pd.DataFrame({"id": [0], "manager_id": [0]})


@pytest.fixture
def dd_employee_df(pd_employee_df):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    return dd.from_pandas(pd_employee_df, npartitions=2)


@pytest.fixture
def spark_employee_df(pd_employee_df):
    ps = pytest.importorskip("pyspark.pandas", reason="Spark not installed, skipping")
    return ps.from_pandas(pd_employee_df)


@pytest.fixture(params=["pd_employee_df", "dd_employee_df", "spark_employee_df"])
def employee_df(request):
    return request.getfixturevalue(request.param)


def test_find_forward_paths_ignores_loops(employee_df):
    dataframes = {"employees": (employee_df, "id")}
    relationships = [("employees", "id", "employees", "manager_id")]
    es = EntitySet(dataframes=dataframes, relationships=relationships)

    paths = list(es.find_forward_paths("employees", "employees"))
    assert len(paths) == 1
    assert paths[0] == []


def test_find_backward_paths(es):
    paths = list(es.find_backward_paths("customers", "log"))
    assert len(paths) == 1

    path = paths[0]

    assert len(path) == 2
    assert path[0]._child_dataframe_name == "sessions"
    assert path[0]._parent_dataframe_name == "customers"
    assert path[1]._child_dataframe_name == "log"
    assert path[1]._parent_dataframe_name == "sessions"


def test_find_backward_paths_multiple_paths(diamond_es):
    paths = list(diamond_es.find_backward_paths("regions", "transactions"))
    assert len(paths) == 2

    path1, path2 = paths

    r1, r2 = path1
    assert r1._child_dataframe_name == "stores"
    assert r1._parent_dataframe_name == "regions"
    assert r2._child_dataframe_name == "transactions"
    assert r2._parent_dataframe_name == "stores"

    r1, r2 = path2
    assert r1._child_dataframe_name == "customers"
    assert r1._parent_dataframe_name == "regions"
    assert r2._child_dataframe_name == "transactions"
    assert r2._parent_dataframe_name == "customers"


def test_find_backward_paths_multiple_relationships(games_es):
    paths = list(games_es.find_backward_paths("teams", "games"))
    assert len(paths) == 2

    path1, path2 = paths
    assert len(path1) == 1
    assert len(path2) == 1
    r1 = path1[0]
    r2 = path2[0]

    assert r1._child_dataframe_name == "games"
    assert r2._child_dataframe_name == "games"
    assert r1._parent_dataframe_name == "teams"
    assert r2._parent_dataframe_name == "teams"

    assert r1._child_column_name == "home_team_id"
    assert r2._child_column_name == "away_team_id"
    assert r1._parent_column_name == "id"
    assert r2._parent_column_name == "id"


def test_has_unique_path(diamond_es):
    assert diamond_es.has_unique_forward_path("customers", "regions")
    assert not diamond_es.has_unique_forward_path("transactions", "regions")


def test_raise_key_error_missing_dataframe(es):
    error_text = "DataFrame testing does not exist in ecommerce"
    with pytest.raises(KeyError, match=error_text):
        es["testing"]

    es_without_id = EntitySet()
    error_text = "DataFrame testing does not exist in entity set"
    with pytest.raises(KeyError, match=error_text):
        es_without_id["testing"]


def test_add_parent_not_index_column(es):
    error_text = "Parent column 'language' is not the index of dataframe régions"
    with pytest.raises(AttributeError, match=error_text):
        es.add_relationship("régions", "language", "customers", "région_id")



from datetime import datetime
from functools import partial

import numpy as np
import pandas as pd
import pandas.api.types as pdtypes

from featuretools.entityset.relationship import RelationshipPath
from featuretools.exceptions import UnknownFeature
from featuretools.feature_base import (
    AggregationFeature,
    DirectFeature,
    GroupByTransformFeature,
    IdentityFeature,
    TransformFeature,
)
from featuretools.utils import Trie
from featuretools.utils.gen_utils import (
    Library,
    get_relationship_column_id,
    import_or_none,
    is_instance,
)

dd = import_or_none("dask.dataframe")
ps = import_or_none("pyspark.pandas")


class FeatureSetCalculator(object):
    """
    Calculates the values of a set of features for given instance ids.
    """

    def __init__(
        self,
        entityset,
        feature_set,
        time_last=None,
        training_window=None,
        precalculated_features=None,
    ):
        """
        Args:
            feature_set (FeatureSet): The features to calculate values for.

            time_last (pd.Timestamp, optional): Last allowed time. Data from exactly this
                time not allowed.

            training_window (Timedelta, optional): Window defining how much time before the cutoff time data
                can be used when calculating features. If None, all data before cutoff time is used.

            precalculated_features (Trie[RelationshipPath -> pd.DataFrame]):
                Maps RelationshipPaths to dataframes of precalculated_features

        """
        self.entityset = entityset
        self.feature_set = feature_set
        self.training_window = training_window

        if time_last is None:
            time_last = datetime.now()

        self.time_last = time_last

        if precalculated_features is None:
            precalculated_features = Trie(path_constructor=RelationshipPath)

        self.precalculated_features = precalculated_features

        # total number of features (including dependencies) to be calculate
        self.num_features = sum(
            len(features1) + len(features2)
            for _, (_, features1, features2) in self.feature_set.feature_trie
        )

    def run(self, instance_ids, progress_callback=None, include_cutoff_time=True):
        """
        Calculate values of features for the given instances of the target
        dataframe.

        Summary of algorithm:
        1. Construct a trie where the edges are relationships and each node
            contains a set of features for a single dataframe. See
            FeatureSet._build_feature_trie.
        2. Initialize a trie for storing dataframes.
        3. Traverse the trie using depth first search. At each node calculate
            the features and store the resulting dataframe in the dataframe
            trie (so that its values can be used by features which depend on
            these features). See _calculate_features_for_dataframe.
        4. Get the dataframe at the root of the trie (for the target dataframe) and
            return the columns corresponding to the requested features.

        Args:
            instance_ids (np.ndarray or pd.Categorical): Instance ids for which
                to build features.

            progress_callback (callable): function to be called with incremental progress updates

            include_cutoff_time (bool): If True, data at cutoff time are included
                in calculating features.

        Returns:
            pd.DataFrame : Pandas DataFrame of calculated feature values.
                Indexed by instance_ids. Columns in same order as features
                passed in.
        """
        assert len(instance_ids) > 0, "0 instance ids provided"

        if progress_callback is None:
            # do nothing for the progress call back if not provided
            def progress_callback(*args):
                pass

        feature_trie = self.feature_set.feature_trie

        df_trie = Trie(path_constructor=RelationshipPath)
        full_dataframe_trie = Trie(path_constructor=RelationshipPath)

        target_dataframe = self.entityset[self.feature_set.target_df_name]

        self._calculate_features_for_dataframe(
            dataframe_name=self.feature_set.target_df_name,
            feature_trie=feature_trie,
            df_trie=df_trie,
            full_dataframe_trie=full_dataframe_trie,
            precalculated_trie=self.precalculated_features,
            filter_column=target_dataframe.ww.index,
            filter_values=instance_ids,
            progress_callback=progress_callback,
            include_cutoff_time=include_cutoff_time,
        )

        # The dataframe for the target dataframe should be stored at the root of
        # df_trie.
        df = df_trie.value

        # Fill in empty rows with default values. This only works for pandas dataframes
        # and is not currently supported for Dask dataframes.
        if isinstance(df, pd.DataFrame):
            index_dtype = df.index.dtype.name
            if df.empty:
                return self.generate_default_df(instance_ids=instance_ids)

            missing_ids = [
                i for i in instance_ids if i not in df[target_dataframe.ww.index]
            ]
            if missing_ids:
                default_df = self.generate_default_df(
                    instance_ids=missing_ids,
                    extra_columns=df.columns,
                )

                df = pd.concat([df, default_df], sort=True)

            df.index.name = self.entityset[self.feature_set.target_df_name].ww.index

            # Order by instance_ids
            unique_instance_ids = pd.unique(instance_ids)
            unique_instance_ids = unique_instance_ids.astype(instance_ids.dtype)
            df = df.reindex(unique_instance_ids)

            # Keep categorical index if original index was categorical
            if index_dtype == "category":
                df.index = df.index.astype("category")

        column_list = []

        for feat in self.feature_set.target_features:
            column_list.extend(feat.get_feature_names())

        if is_instance(df, (dd, ps), "DataFrame"):
            column_list.extend([target_dataframe.ww.index])

        return df[column_list]

    def _calculate_features_for_dataframe(
        self,
        dataframe_name,
        feature_trie,
        df_trie,
        full_dataframe_trie,
        precalculated_trie,
        filter_column,
        filter_values,
        parent_data=None,
        progress_callback=None,
        include_cutoff_time=True,
    ):
        """
        Generate dataframes with features calculated for this node of the trie,
        and all descendant nodes. The dataframes will be stored in df_trie.

        Args:
            dataframe_name (str): The name of the dataframe to calculate features for.

            feature_trie (Trie): the trie with sets of features to calculate.
                The root contains features for the given dataframe.

            df_trie (Trie): a parallel trie for storing dataframes. The
                dataframe with features calculated will be placed in the root.

            full_dataframe_trie (Trie): a trie storing dataframes will all dataframe
                rows, for features that are uses_full_dataframe.

            precalculated_trie (Trie): a parallel trie containing dataframes
                with precalculated features. The dataframe specified by dataframe_name
                will be at the root.

            filter_column (str): The name of the column to filter this
                dataframe by.

            filter_values (pd.Series): The values to filter the filter_column
                to.

            parent_data (tuple[Relationship, list[str], pd.DataFrame]): Data
                related to the parent of this trie. This will only be present if
                the relationship points from this dataframe to the parent dataframe. A
                3 tuple of (parent_relationship,
                ancestor_relationship_columns, parent_df).
                ancestor_relationship_columns is the names of columns which
                link the parent dataframe to its ancestors.

            include_cutoff_time (bool): If True, data at cutoff time are included
                in calculating features.

        """
        # Step 1: Get a dataframe for the given dataframe name, filtered by the given
        # conditions.

        (
            need_full_dataframe,
            full_dataframe_features,
            not_full_dataframe_features,
        ) = feature_trie.value

        all_features = full_dataframe_features | not_full_dataframe_features
        columns = self._necessary_columns(dataframe_name, all_features)

        # If we need the full dataframe then don't filter by filter_values.
        if need_full_dataframe:
            query_column = None
            query_values = None
        else:
            query_column = filter_column
            query_values = filter_values

        df = self.entityset.query_by_values(
            dataframe_name=dataframe_name,
            instance_vals=query_values,
            column_name=query_column,
            columns=columns,
            time_last=self.time_last,
            training_window=self.training_window,
            include_cutoff_time=include_cutoff_time,
        )

        # call to update timer
        progress_callback(0)

        # Step 2: Add columns to the dataframe linking it to all ancestors.
        new_ancestor_relationship_columns = []
        if parent_data:
            parent_relationship, ancestor_relationship_columns, parent_df = parent_data

            if ancestor_relationship_columns:
                (
                    df,
                    new_ancestor_relationship_columns,
                ) = self._add_ancestor_relationship_columns(
                    df,
                    parent_df,
                    ancestor_relationship_columns,
                    parent_relationship,
                )

            # Add the column linking this dataframe to its parent, so that
            # descendants get linked to the parent.
            new_ancestor_relationship_columns.append(
                parent_relationship._child_column_name,
            )

        # call to update timer
        progress_callback(0)

        # Step 3: Recurse on children.

        # Pass filtered values, even if we are using a full df.
        if need_full_dataframe:
            if is_instance(filter_values, dd, "Series"):
                msg = "Cannot use primitives that require full dataframe with Dask EntitySets"
                raise ValueError(msg)
            filtered_df = df[df[filter_column].isin(filter_values)]
        else:
            filtered_df = df

        for edge, sub_trie in feature_trie.children():
            is_forward, relationship = edge
            if is_forward:
                sub_dataframe_name = relationship.parent_dataframe.ww.name
                sub_filter_column = relationship._parent_column_name
                sub_filter_values = filtered_df[relationship._child_column_name]
                parent_data = None
            else:
                sub_dataframe_name = relationship.child_dataframe.ww.name
                sub_filter_column = relationship._child_column_name
                sub_filter_values = filtered_df[relationship._parent_column_name]

                parent_data = (relationship, new_ancestor_relationship_columns, df)

            sub_df_trie = df_trie.get_node([edge])
            sub_full_dataframe_trie = full_dataframe_trie.get_node([edge])
            sub_precalc_trie = precalculated_trie.get_node([edge])
            self._calculate_features_for_dataframe(
                dataframe_name=sub_dataframe_name,
                feature_trie=sub_trie,
                df_trie=sub_df_trie,
                full_dataframe_trie=sub_full_dataframe_trie,
                precalculated_trie=sub_precalc_trie,
                filter_column=sub_filter_column,
                filter_values=sub_filter_values,
                parent_data=parent_data,
                progress_callback=progress_callback,
                include_cutoff_time=include_cutoff_time,
            )

        # Step 4: Calculate the features for this dataframe.
        #
        # All dependencies of the features for this dataframe have been calculated
        # by the above recursive calls, and their results stored in df_trie.

        # Add any precalculated features.
        precalculated_features_df = precalculated_trie.value
        if precalculated_features_df is not None:
            # Left outer merge to keep all rows of df.
            df = df.merge(
                precalculated_features_df,
                how="left",
                left_index=True,
                right_index=True,
                suffixes=("", "_precalculated"),
            )

        # call to update timer
        progress_callback(0)

        # First, calculate any features that require the full dataframe. These can
        # be calculated first because all of their dependents are included in
        # full_dataframe_features.
        if need_full_dataframe:
            df = self._calculate_features(
                df,
                full_dataframe_trie,
                full_dataframe_features,
                progress_callback,
            )

            # Store full dataframe
            full_dataframe_trie.value = df

            # Filter df so that features that don't require the full dataframe are
            # only calculated on the necessary instances.
            df = df[df[filter_column].isin(filter_values)]

        # Calculate all features that don't require the full dataframe.
        df = self._calculate_features(
            df,
            df_trie,
            not_full_dataframe_features,
            progress_callback,
        )

        # Step 5: Store the dataframe for this dataframe at the root of df_trie, so
        # that it can be accessed by the caller.
        df_trie.value = df

    def _calculate_features(self, df, df_trie, features, progress_callback):
        # Group the features so that each group can be calculated together.
        # The groups must also be in topological order (if A is a transform of B
        # then B must be in a group before A).
        feature_groups = self.feature_set.group_features(features)

        for group in feature_groups:
            representative_feature = group[0]
            handler = self._feature_type_handler(representative_feature)
            df = handler(group, df, df_trie, progress_callback)

        return df

    def _add_ancestor_relationship_columns(
        self,
        child_df,
        parent_df,
        ancestor_relationship_columns,
        relationship,
    ):
        """
        Merge ancestor_relationship_columns from parent_df into child_df, adding a prefix to
        each column name specifying the relationship.

        Return the updated df and the new relationship column names.

        Args:
            child_df (pd.DataFrame): The dataframe to add relationship columns to.
            parent_df (pd.DataFrame): The dataframe to copy relationship columns from.
            ancestor_relationship_columns (list[str]): The names of
                relationship columns in the parent_df to copy into child_df.
            relationship (Relationship): the relationship through which the
                child is connected to the parent.
        """
        relationship_name = relationship.parent_name
        new_relationship_columns = [
            "%s.%s" % (relationship_name, col) for col in ancestor_relationship_columns
        ]

        # create an intermediate dataframe which shares a column
        # with the child dataframe and has a column with the
        # original parent's id.
        col_map = {relationship._parent_column_name: relationship._child_column_name}
        for child_column, parent_column in zip(
            new_relationship_columns,
            ancestor_relationship_columns,
        ):
            col_map[parent_column] = child_column

        merge_df = parent_df[list(col_map.keys())].rename(columns=col_map)

        merge_df.index.name = None  # change index name for merge

        # Merge the dataframe, adding the relationship columns to the child.
        # Left outer join so that all rows in child are kept (if it contains
        # all rows of the dataframe then there may not be corresponding rows in the
        # parent_df).
        df = child_df.merge(
            merge_df,
            how="left",
            left_on=relationship._child_column_name,
            right_on=relationship._child_column_name,
        )

        # ensure index is maintained
        # TODO: Review for dask dataframes
        if isinstance(df, pd.DataFrame):
            df.set_index(
                relationship.child_dataframe.ww.index,
                drop=False,
                inplace=True,
            )

        return df, new_relationship_columns

    def generate_default_df(self, instance_ids, extra_columns=None):
        default_row = []
        default_cols = []
        for f in self.feature_set.target_features:
            for name in f.get_feature_names():
                default_cols.append(name)
                default_row.append(f.default_value)

        default_matrix = [default_row] * len(instance_ids)
        default_df = pd.DataFrame(
            default_matrix,
            columns=default_cols,
            index=instance_ids,
            dtype="object",
        )
        index_name = self.entityset[self.feature_set.target_df_name].ww.index
        default_df.index.name = index_name
        if extra_columns is not None:
            for c in extra_columns:
                if c not in default_df.columns:
                    default_df[c] = [np.nan] * len(instance_ids)
        return default_df

    def _feature_type_handler(self, f):
        if type(f) == TransformFeature:
            return self._calculate_transform_features
        elif type(f) == GroupByTransformFeature:
            return self._calculate_groupby_features
        elif type(f) == DirectFeature:
            return self._calculate_direct_features
        elif type(f) == AggregationFeature:
            return self._calculate_agg_features
        elif type(f) == IdentityFeature:
            return self._calculate_identity_features
        else:
            raise UnknownFeature("{} feature unknown".format(f.__class__))

    def _calculate_identity_features(self, features, df, _df_trie, progress_callback):
        for f in features:
            assert f.get_name() in df.columns, (
                'Column "%s" missing frome dataframe' % f.get_name()
            )

        progress_callback(len(features) / float(self.num_features))

        return df

    def _calculate_transform_features(
        self,
        features,
        frame,
        _df_trie,
        progress_callback,
    ):
        frame_empty = frame.empty if isinstance(frame, pd.DataFrame) else False
        feature_values = []
        for f in features:
            # handle when no data
            if frame_empty:
                # Even though we are adding the default values here, when these new
                # features are added to the dataframe in update_feature_columns, they
                # are added as empty columns since the dataframe itself is empty.
                feature_values.append(
                    (f, [f.default_value for _ in range(f.number_output_features)]),
                )
                progress_callback(1 / float(self.num_features))
                continue

            # collect only the columns we need for this transformation

            column_data = [frame[bf.get_name()] for bf in f.base_features]

            feature_func = f.get_function()
            # apply the function to the relevant dataframe slice and add the
            # feature row to the results dataframe.
            if f.primitive.uses_calc_time:
                values = feature_func(*column_data, time=self.time_last)
            else:
                values = feature_func(*column_data)

            # if we don't get just the values, the assignment breaks when indexes don't match
            if f.number_output_features > 1:
                values = [strip_values_if_series(value) for value in values]
            else:
                values = [strip_values_if_series(values)]

            feature_values.append((f, values))

            progress_callback(1 / float(self.num_features))

        frame = update_feature_columns(feature_values, frame)
        return frame

    def _calculate_groupby_features(self, features, frame, _df_trie, progress_callback):
        # set default values to handle the null group
        default_values = {}
        for f in features:
            for name in f.get_feature_names():
                default_values[name] = f.default_value

        frame = pd.concat(
            [frame, pd.DataFrame(default_values, index=frame.index)],
            axis=1,
        )

        # handle when no data
        if frame.shape[0] == 0:
            progress_callback(len(features) / float(self.num_features))

            return frame

        groupby = features[0].groupby.get_name()
        grouped = frame.groupby(groupby)
        groups = frame[
            groupby
        ].unique()  # get all the unique group name to iterate over later

        for f in features:
            feature_vals = []
            for _ in range(f.number_output_features):
                feature_vals.append([])

            for group in groups:
                # skip null key if it exists
                if pd.isnull(group):
                    continue

                column_names = [bf.get_name() for bf in f.base_features]
                # exclude the groupby column from being passed to the function
                column_data = [
                    grouped[name].get_group(group) for name in column_names[:-1]
                ]
                feature_func = f.get_function()

                # apply the function to the relevant dataframe slice and add the
                # feature row to the results dataframe.
                if f.primitive.uses_calc_time:
                    values = feature_func(*column_data, time=self.time_last)
                else:
                    values = feature_func(*column_data)

                if f.number_output_features == 1:
                    values = [values]

                # make sure index is aligned
                for i, value in enumerate(values):
                    if isinstance(value, pd.Series):
                        value.index = column_data[0].index
                    else:
                        value = pd.Series(value, index=column_data[0].index)
                    feature_vals[i].append(value)

            if any(feature_vals):
                assert len(feature_vals) == len(f.get_feature_names())
                for col_vals, name in zip(feature_vals, f.get_feature_names()):
                    frame[name].update(pd.concat(col_vals))

            progress_callback(1 / float(self.num_features))

        return frame

    def _calculate_direct_features(
        self,
        features,
        child_df,
        df_trie,
        progress_callback,
    ):
        path = features[0].relationship_path
        assert len(path) == 1, "Error calculating DirectFeatures, len(path) != 1"

        parent_df = df_trie.get_node([path[0]]).value
        _is_forward, relationship = path[0]
        merge_col = relationship._child_column_name

        # generate a mapping of old column names (in the parent dataframe) to
        # new column names (in the child dataframe) for the merge
        col_map = {relationship._parent_column_name: merge_col}
        index_as_feature = None

        fillna_dict = {}
        for f in features:
            feature_defaults = {
                name: f.default_value
                for name in f.get_feature_names()
                if not pd.isna(f.default_value)
            }
            fillna_dict.update(feature_defaults)
            if f.base_features[0].get_name() == relationship._parent_column_name:
                index_as_feature = f
            base_names = f.base_features[0].get_feature_names()
            for name, base_name in zip(f.get_feature_names(), base_names):
                if name in child_df.columns:
                    continue
                col_map[base_name] = name

        # merge the identity feature from the parent dataframe into the child
        merge_df = parent_df[list(col_map.keys())].rename(columns=col_map)
        if is_instance(merge_df, (dd, ps), "DataFrame"):
            new_df = child_df.merge(
                merge_df,
                left_on=merge_col,
                right_on=merge_col,
                how="left",
            )
        else:
            if index_as_feature is not None:
                merge_df.set_index(
                    index_as_feature.get_name(),
                    inplace=True,
                    drop=False,
                )
            else:
                merge_df.set_index(merge_col, inplace=True)

            new_df = child_df.merge(
                merge_df,
                left_on=merge_col,
                right_index=True,
                how="left",
            )

        progress_callback(len(features) / float(self.num_features))

        return new_df.fillna(fillna_dict)

    def _calculate_agg_features(self, features, frame, df_trie, progress_callback):
        test_feature = features[0]
        child_dataframe = test_feature.base_features[0].dataframe
        base_frame = df_trie.get_node(test_feature.relationship_path).value
        parent_merge_col = test_feature.relationship_path[0][1]._parent_column_name
        # Sometimes approximate features get computed in a previous filter frame
        # and put in the current one dynamically,
        # so there may be existing features here
        fl = []
        for f in features:
            for ind in f.get_feature_names():
                if ind not in frame.columns:
                    fl.append(f)
                    break
        features = fl
        if not len(features):
            progress_callback(len(features) / float(self.num_features))
            return frame

        # handle where
        base_frame_empty = (
            base_frame.empty if isinstance(base_frame, pd.DataFrame) else False
        )
        where = test_feature.where
        if where is not None and not base_frame_empty:
            base_frame = base_frame.loc[base_frame[where.get_name()]]

        # when no child data, just add all the features to frame with nan
        base_frame_empty = (
            base_frame.empty if isinstance(base_frame, pd.DataFrame) else False
        )
        if base_frame_empty:
            feature_values = []
            for f in features:
                feature_values.append((f, np.full(f.number_output_features, np.nan)))
                progress_callback(1 / float(self.num_features))
            frame = update_feature_columns(feature_values, frame)
        else:
            relationship_path = test_feature.relationship_path

            groupby_col = get_relationship_column_id(relationship_path)

            # if the use_previous property exists on this feature, include only the
            # instances from the child dataframe included in that Timedelta
            use_previous = test_feature.use_previous
            if use_previous:
                # Filter by use_previous values
                time_last = self.time_last
                if use_previous.has_no_observations():
                    time_first = time_last - use_previous
                    ti = child_dataframe.ww.time_index
                    if ti is not None:
                        base_frame = base_frame[base_frame[ti] >= time_first]
                else:
                    n = use_previous.get_value("o")

                    def last_n(df):
                        return df.iloc[-n:]

                    base_frame = base_frame.groupby(
                        groupby_col,
                        observed=True,
                        sort=False,
                        group_keys=False,
                    ).apply(last_n)

            to_agg = {}
            agg_rename = {}
            to_apply = set()
            # apply multi-column and time-dependent features as we find them, and
            # save aggregable features for later
            for f in features:
                if _can_agg(f):
                    column_id = f.base_features[0].get_name()
                    if column_id not in to_agg:
                        to_agg[column_id] = []
                    if is_instance(base_frame, dd, "DataFrame"):
                        func = f.get_function(agg_type=Library.DASK)
                    elif is_instance(base_frame, ps, "DataFrame"):
                        func = f.get_function(agg_type=Library.SPARK)
                    else:
                        func = f.get_function()

                    # for some reason, using the string count is significantly
                    # faster than any method a primitive can return
                    # https://stackoverflow.com/questions/55731149/use-a-function-instead-of-string-in-pandas-groupby-agg
                    if func == pd.Series.count:
                        func = "count"

                    funcname = func
                    if callable(func):
                        # if the same function is being applied to the same
                        # column twice, wrap it in a partial to avoid
                        # duplicate functions
                        funcname = str(id(func))
                        if "{}-{}".format(column_id, funcname) in agg_rename:
                            func = partial(func)
                            funcname = str(id(func))

                        func.__name__ = funcname

                    if dd and isinstance(func, dd.Aggregation):
                        # TODO: handle aggregation being applied to same column twice
                        # (see above partial wrapping of functions)
                        funcname = func.__name__

                    to_agg[column_id].append(func)
                    # this is used below to rename columns that pandas names for us
                    agg_rename["{}-{}".format(column_id, funcname)] = f.get_name()
                    continue

                to_apply.add(f)

            # Apply the non-aggregable functions generate a new dataframe, and merge
            # it with the existing one
            if len(to_apply):
                wrap = agg_wrapper(to_apply, self.time_last)
                # groupby_col can be both the name of the index and a column,
                # to silence pandas warning about ambiguity we explicitly pass
                # the column (in actuality grouping by both index and group would
                # work)
                to_merge = base_frame.groupby(
                    base_frame[groupby_col],
                    observed=True,
                    sort=False,
                    group_keys=False,
                ).apply(wrap)
                frame = pd.merge(
                    left=frame,
                    right=to_merge,
                    left_index=True,
                    right_index=True,
                    how="left",
                )

                progress_callback(len(to_apply) / float(self.num_features))

            # Apply the aggregate functions to generate a new dataframe, and merge
            # it with the existing one
            if len(to_agg):
                # groupby_col can be both the name of the index and a column,
                # to silence pandas warning about ambiguity we explicitly pass
                # the column (in actuality grouping by both index and group would
                # work)
                if is_instance(base_frame, (dd, ps), "DataFrame"):
                    to_merge = base_frame.groupby(groupby_col).agg(to_agg)

                else:
                    to_merge = base_frame.groupby(
                        base_frame[groupby_col],
                        observed=True,
                        sort=False,
                    ).agg(to_agg)
                # rename columns to the correct feature names
                to_merge.columns = [agg_rename["-".join(x)] for x in to_merge.columns]
                to_merge = to_merge[list(agg_rename.values())]

                # Workaround for pandas bug where categories are in the wrong order
                # see: https://github.com/pandas-dev/pandas/issues/22501
                #
                # Pandas claims that bug is fixed but it still shows up in some
                # cases.  More investigation needed.
                if pdtypes.is_categorical_dtype(frame.index):
                    categories = pdtypes.CategoricalDtype(
                        categories=frame.index.categories,
                    )
                    to_merge.index = to_merge.index.astype(object).astype(categories)

                if is_instance(frame, (dd, ps), "DataFrame"):
                    frame = frame.merge(
                        to_merge,
                        left_on=parent_merge_col,
                        right_index=True,
                        how="left",
                    )
                else:
                    frame = pd.merge(
                        left=frame,
                        right=to_merge,
                        left_index=True,
                        right_index=True,
                        how="left",
                    )

                # determine number of features that were just merged
                progress_callback(len(to_merge.columns) / float(self.num_features))

        # Handle default values
        fillna_dict = {}
        for f in features:
            feature_defaults = {name: f.default_value for name in f.get_feature_names()}
            fillna_dict.update(feature_defaults)

        frame = frame.fillna(fillna_dict)

        return frame

    def _necessary_columns(self, dataframe_name, feature_names):
        # We have to keep all index and foreign columns because we don't know what forward
        # relationships will come from this node.
        df = self.entityset[dataframe_name]
        index_columns = {
            col
            for col in df.columns
            if {"index", "foreign_key", "time_index"} & df.ww.semantic_tags[col]
        }
        features = (self.feature_set.features_by_name[name] for name in feature_names)

        feature_columns = {
            f.column_name for f in features if isinstance(f, IdentityFeature)
        }
        return list(index_columns | feature_columns)


def _can_agg(feature):
    assert isinstance(feature, AggregationFeature)
    base_features = feature.base_features
    if feature.where is not None:
        base_features = [
            bf.get_name()
            for bf in base_features
            if bf.get_name() != feature.where.get_name()
        ]

    if feature.primitive.uses_calc_time:
        return False
    single_output = feature.primitive.number_output_features == 1
    return len(base_features) == 1 and single_output


def agg_wrapper(feats, time_last):
    def wrap(df):
        d = {}
        feature_values = []
        for f in feats:
            func = f.get_function()
            column_ids = [bf.get_name() for bf in f.base_features]
            args = [df[v] for v in column_ids]

            if f.primitive.uses_calc_time:
                values = func(*args, time=time_last)
            else:
                values = func(*args)

            if f.number_output_features == 1:
                values = [values]
            feature_values.append((f, values))

        d = update_feature_columns(feature_values, d)

        return pd.Series(d)

    return wrap


def update_feature_columns(feature_data, data):
    new_cols = {}
    for item in feature_data:
        names = item[0].get_feature_names()
        values = item[1]
        assert len(names) == len(values)
        for name, value in zip(names, values):
            new_cols[name] = value

    # Handle the case where a dict is being updated
    if isinstance(data, dict):
        data.update(new_cols)
        return data

    # Handle pandas input
    if isinstance(data, pd.DataFrame):
        return pd.concat([data, pd.DataFrame(new_cols, index=data.index)], axis=1)

    # Handle dask/spark input
    for name, col in new_cols.items():
        col.name = name
        if is_instance(data, dd, "DataFrame"):
            data = dd.concat([data, col], axis=1)
        else:
            data = ps.concat([data, col], axis=1)
    return data


def strip_values_if_series(values):
    if isinstance(values, pd.Series):
        values = values.values
    return values



from string import punctuation

DELIMITERS = "[ \n\t]"
PUNCTUATION_AND_WHITESPACE = f"[{punctuation}\n\t ]"

common_words_1000 = frozenset(
    [
        "the",
        "of",
        "to",
        "and",
        "a",
        "in",
        "is",
        "it",
        "you",
        "that",
        "he",
        "was",
        "for",
        "on",
        "are",
        "with",
        "as",
        "i",
        "his",
        "they",
        "be",
        "at",
        "one",
        "have",
        "this",
        "from",
        "or",
        "had",
        "by",
        "not",
        "word",
        "but",
        "what",
        "some",
        "we",
        "can",
        "out",
        "other",
        "were",
        "all",
        "there",
        "when",
        "up",
        "use",
        "your",
        "how",
        "said",
        "an",
        "each",
        "she",
        "which",
        "do",
        "their",
        "time",
        "if",
        "will",
        "way",
        "about",
        "many",
        "then",
        "them",
        "write",
        "would",
        "like",
        "so",
        "these",
        "her",
        "long",
        "make",
        "thing",
        "see",
        "him",
        "two",
        "has",
        "look",
        "more",
        "day",
        "could",
        "go",
        "come",
        "did",
        "number",
        "sound",
        "no",
        "most",
        "people",
        "my",
        "over",
        "know",
        "water",
        "than",
        "call",
        "first",
        "who",
        "may",
        "down",
        "side",
        "been",
        "now",
        "find",
        "any",
        "new",
        "work",
        "part",
        "take",
        "get",
        "place",
        "made",
        "live",
        "where",
        "after",
        "back",
        "little",
        "only",
        "round",
        "man",
        "year",
        "came",
        "show",
        "every",
        "good",
        "me",
        "give",
        "our",
        "under",
        "name",
        "very",
        "through",
        "just",
        "form",
        "sentence",
        "great",
        "think",
        "say",
        "help",
        "low",
        "line",
        "differ",
        "turn",
        "cause",
        "much",
        "mean",
        "before",
        "move",
        "right",
        "boy",
        "old",
        "too",
        "same",
        "tell",
        "does",
        "set",
        "three",
        "want",
        "air",
        "well",
        "also",
        "play",
        "small",
        "end",
        "put",
        "home",
        "read",
        "hand",
        "port",
        "large",
        "spell",
        "add",
        "even",
        "land",
        "here",
        "must",
        "big",
        "high",
        "such",
        "follow",
        "act",
        "why",
        "ask",
        "men",
        "change",
        "went",
        "light",
        "kind",
        "off",
        "need",
        "house",
        "picture",
        "try",
        "us",
        "again",
        "animal",
        "point",
        "mother",
        "world",
        "near",
        "build",
        "self",
        "earth",
        "father",
        "head",
        "stand",
        "own",
        "page",
        "should",
        "country",
        "found",
        "answer",
        "school",
        "grow",
        "study",
        "still",
        "learn",
        "plant",
        "cover",
        "food",
        "sun",
        "four",
        "between",
        "state",
        "keep",
        "eye",
        "never",
        "last",
        "let",
        "thought",
        "city",
        "tree",
        "cross",
        "farm",
        "hard",
        "start",
        "might",
        "story",
        "saw",
        "far",
        "sea",
        "draw",
        "left",
        "late",
        "run",
        "don't",
        "while",
        "press",
        "close",
        "night",
        "real",
        "life",
        "few",
        "north",
        "open",
        "seem",
        "together",
        "next",
        "white",
        "children",
        "begin",
        "got",
        "walk",
        "example",
        "ease",
        "paper",
        "group",
        "always",
        "music",
        "those",
        "both",
        "mark",
        "often",
        "letter",
        "until",
        "mile",
        "river",
        "car",
        "feet",
        "care",
        "second",
        "book",
        "carry",
        "took",
        "science",
        "eat",
        "room",
        "friend",
        "began",
        "idea",
        "fish",
        "mountain",
        "stop",
        "once",
        "base",
        "hear",
        "horse",
        "cut",
        "sure",
        "watch",
        "color",
        "face",
        "wood",
        "main",
        "enough",
        "plain",
        "girl",
        "usual",
        "young",
        "ready",
        "above",
        "ever",
        "red",
        "list",
        "though",
        "feel",
        "talk",
        "bird",
        "soon",
        "body",
        "dog",
        "family",
        "direct",
        "pose",
        "leave",
        "song",
        "measure",
        "door",
        "product",
        "black",
        "short",
        "numeral",
        "class",
        "wind",
        "question",
        "happen",
        "complete",
        "ship",
        "area",
        "half",
        "rock",
        "order",
        "fire",
        "south",
        "problem",
        "piece",
        "told",
        "knew",
        "pass",
        "since",
        "top",
        "whole",
        "king",
        "space",
        "heard",
        "best",
        "hour",
        "better",
        "true",
        "during",
        "hundred",
        "five",
        "remember",
        "step",
        "early",
        "hold",
        "west",
        "ground",
        "interest",
        "reach",
        "fast",
        "verb",
        "sing",
        "listen",
        "six",
        "table",
        "travel",
        "less",
        "morning",
        "ten",
        "simple",
        "several",
        "vowel",
        "toward",
        "war",
        "lay",
        "against",
        "pattern",
        "slow",
        "center",
        "love",
        "person",
        "money",
        "serve",
        "appear",
        "road",
        "map",
        "rain",
        "rule",
        "govern",
        "pull",
        "cold",
        "notice",
        "voice",
        "unit",
        "power",
        "town",
        "fine",
        "certain",
        "fly",
        "fall",
        "lead",
        "cry",
        "dark",
        "machine",
        "note",
        "wait",
        "plan",
        "figure",
        "star",
        "box",
        "noun",
        "field",
        "rest",
        "correct",
        "able",
        "pound",
        "done",
        "beauty",
        "drive",
        "stood",
        "contain",
        "front",
        "teach",
        "week",
        "final",
        "gave",
        "green",
        "oh",
        "quick",
        "develop",
        "ocean",
        "warm",
        "free",
        "minute",
        "strong",
        "special",
        "mind",
        "behind",
        "clear",
        "tail",
        "produce",
        "fact",
        "street",
        "inch",
        "multiply",
        "nothing",
        "course",
        "stay",
        "wheel",
        "full",
        "force",
        "blue",
        "object",
        "decide",
        "surface",
        "deep",
        "moon",
        "island",
        "foot",
        "system",
        "busy",
        "test",
        "record",
        "boat",
        "common",
        "gold",
        "possible",
        "plane",
        "stead",
        "dry",
        "wonder",
        "laugh",
        "thousand",
        "ago",
        "ran",
        "check",
        "game",
        "shape",
        "equate",
        "hot",
        "miss",
        "brought",
        "heat",
        "snow",
        "tire",
        "bring",
        "yes",
        "distant",
        "fill",
        "east",
        "paint",
        "language",
        "among",
        "grand",
        "ball",
        "yet",
        "wave",
        "drop",
        "heart",
        "am",
        "present",
        "heavy",
        "dance",
        "engine",
        "position",
        "arm",
        "wide",
        "sail",
        "material",
        "size",
        "vary",
        "settle",
        "speak",
        "weight",
        "general",
        "ice",
        "matter",
        "circle",
        "pair",
        "include",
        "divide",
        "syllable",
        "felt",
        "perhaps",
        "pick",
        "sudden",
        "count",
        "square",
        "reason",
        "length",
        "represent",
        "art",
        "subject",
        "region",
        "energy",
        "hunt",
        "probable",
        "bed",
        "brother",
        "egg",
        "ride",
        "cell",
        "believe",
        "fraction",
        "forest",
        "sit",
        "race",
        "window",
        "store",
        "summer",
        "train",
        "sleep",
        "prove",
        "lone",
        "leg",
        "exercise",
        "wall",
        "catch",
        "mount",
        "wish",
        "sky",
        "board",
        "joy",
        "winter",
        "sat",
        "written",
        "wild",
        "instrument",
        "kept",
        "glass",
        "grass",
        "cow",
        "job",
        "edge",
        "sign",
        "visit",
        "past",
        "soft",
        "fun",
        "bright",
        "gas",
        "weather",
        "month",
        "million",
        "bear",
        "finish",
        "happy",
        "hope",
        "flower",
        "clothe",
        "strange",
        "gone",
        "jump",
        "baby",
        "eight",
        "village",
        "meet",
        "root",
        "buy",
        "raise",
        "solve",
        "metal",
        "whether",
        "push",
        "seven",
        "paragraph",
        "third",
        "shall",
        "held",
        "hair",
        "describe",
        "cook",
        "floor",
        "either",
        "result",
        "burn",
        "hill",
        "safe",
        "cat",
        "century",
        "consider",
        "type",
        "law",
        "bit",
        "coast",
        "copy",
        "phrase",
        "silent",
        "tall",
        "sand",
        "soil",
        "roll",
        "temperature",
        "finger",
        "industry",
        "value",
        "fight",
        "lie",
        "beat",
        "excite",
        "natural",
        "view",
        "sense",
        "ear",
        "else",
        "quite",
        "broke",
        "case",
        "middle",
        "kill",
        "son",
        "lake",
        "moment",
        "scale",
        "loud",
        "spring",
        "observe",
        "child",
        "straight",
        "consonant",
        "nation",
        "dictionary",
        "milk",
        "speed",
        "method",
        "organ",
        "pay",
        "age",
        "section",
        "dress",
        "cloud",
        "surprise",
        "quiet",
        "stone",
        "tiny",
        "climb",
        "cool",
        "design",
        "poor",
        "lot",
        "experiment",
        "bottom",
        "key",
        "iron",
        "single",
        "stick",
        "flat",
        "twenty",
        "skin",
        "smile",
        "crease",
        "hole",
        "trade",
        "melody",
        "trip",
        "office",
        "receive",
        "row",
        "mouth",
        "exact",
        "symbol",
        "die",
        "least",
        "trouble",
        "shout",
        "except",
        "wrote",
        "seed",
        "tone",
        "join",
        "suggest",
        "clean",
        "break",
        "lady",
        "yard",
        "rise",
        "bad",
        "blow",
        "oil",
        "blood",
        "touch",
        "grew",
        "cent",
        "mix",
        "team",
        "wire",
        "cost",
        "lost",
        "brown",
        "wear",
        "garden",
        "equal",
        "sent",
        "choose",
        "fell",
        "fit",
        "flow",
        "fair",
        "bank",
        "collect",
        "save",
        "control",
        "decimal",
        "gentle",
        "woman",
        "captain",
        "practice",
        "separate",
        "difficult",
        "doctor",
        "please",
        "protect",
        "noon",
        "whose",
        "locate",
        "ring",
        "character",
        "insect",
        "caught",
        "period",
        "indicate",
        "radio",
        "spoke",
        "atom",
        "human",
        "history",
        "effect",
        "electric",
        "expect",
        "crop",
        "modern",
        "element",
        "hit",
        "student",
        "corner",
        "party",
        "supply",
        "bone",
        "rail",
        "imagine",
        "provide",
        "agree",
        "thus",
        "capital",
        "won't",
        "chair",
        "danger",
        "fruit",
        "rich",
        "thick",
        "soldier",
        "process",
        "operate",
        "guess",
        "necessary",
        "sharp",
        "wing",
        "create",
        "neighbor",
        "wash",
        "bat",
        "rather",
        "crowd",
        "corn",
        "compare",
        "poem",
        "string",
        "bell",
        "depend",
        "meat",
        "rub",
        "tube",
        "famous",
        "dollar",
        "stream",
        "fear",
        "sight",
        "thin",
        "triangle",
        "planet",
        "hurry",
        "chief",
        "colony",
        "clock",
        "mine",
        "tie",
        "enter",
        "major",
        "fresh",
        "search",
        "send",
        "yellow",
        "gun",
        "allow",
        "print",
        "dead",
        "spot",
        "desert",
        "suit",
        "current",
        "lift",
        "rose",
        "continue",
        "block",
        "chart",
        "hat",
        "sell",
        "success",
        "company",
        "subtract",
        "event",
        "particular",
        "deal",
        "swim",
        "term",
        "opposite",
        "wife",
        "shoe",
        "shoulder",
        "spread",
        "arrange",
        "camp",
        "invent",
        "cotton",
        "born",
        "determine",
        "quart",
        "nine",
        "truck",
        "noise",
        "level",
        "chance",
        "gather",
        "shop",
        "stretch",
        "throw",
        "shine",
        "property",
        "column",
        "molecule",
        "select",
        "wrong",
        "gray",
        "repeat",
        "require",
        "broad",
        "prepare",
        "salt",
        "nose",
        "plural",
        "anger",
        "claim",
        "continent",
        "oxygen",
        "sugar",
        "death",
        "pretty",
        "skill",
        "women",
        "season",
        "solution",
        "magnet",
        "silver",
        "thank",
        "branch",
        "match",
        "suffix",
        "especially",
        "fig",
        "afraid",
        "huge",
        "sister",
        "steel",
        "discuss",
        "forward",
        "similar",
        "guide",
        "experience",
        "score",
        "apple",
        "bought",
        "led",
        "pitch",
        "coat",
        "mass",
        "card",
        "band",
        "rope",
        "slip",
        "win",
        "dream",
        "evening",
        "condition",
        "feed",
        "tool",
        "total",
        "basic",
        "smell",
        "valley",
        "nor",
        "double",
        "seat",
        "arrive",
        "master",
        "track",
        "parent",
        "shore",
        "division",
        "sheet",
        "substance",
        "favor",
        "connect",
        "post",
        "spend",
        "chord",
        "fat",
        "glad",
        "original",
        "share",
        "station",
        "dad",
        "bread",
        "charge",
        "proper",
        "bar",
        "offer",
        "segment",
        "slave",
        "duck",
        "instant",
        "market",
        "degree",
        "populate",
        "chick",
        "dear",
        "enemy",
        "reply",
        "drink",
        "occur",
        "support",
        "speech",
        "nature",
        "range",
        "steam",
        "motion",
        "path",
        "liquid",
        "log",
        "meant",
        "quotient",
        "teeth",
        "shell",
        "neck",
    ],
)  # https://gist.github.com/deekayen/4148741



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import BooleanNullable, Datetime

from featuretools.primitives.base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class IsYearEnd(TransformPrimitive):
    """Determines if a date falls on the end of a year.

    Examples:
        >>> import numpy as np
        >>> from datetime import datetime
        >>> dates = [datetime(2019, 12, 31),
        ...          datetime(2019, 1, 1),
        ...          datetime(2019, 11, 30),
        ...          np.nan]
        >>> is_year_end = IsYearEnd()
        >>> is_year_end(dates).tolist()
        [True, False, False, False]
    """

    name = "is_year_end"
    input_types = [ColumnSchema(logical_type=Datetime)]
    return_type = ColumnSchema(logical_type=BooleanNullable)
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "whether {} occurred on the end of a year"

    def get_function(self):
        def is_year_end(vals):
            return vals.dt.is_year_end

        return is_year_end



import numpy as np
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.utils.gen_utils import Library


class SubtractNumeric(TransformPrimitive):
    """Performs element-wise subtraction of two lists.

    Description:
        Given a list of values X and a list of values
        Y, determine the difference of each value
        in X from its corresponding value in Y.

    Args:
        commutative (bool): determines if Deep Feature Synthesis should
            generate both x - y and y - x, or just one. If True, there is no
            guarantee which of the two will be generated. Defaults to True.

    Notes:
        commutative is True by default since False would result in 2 perfectly
        correlated series.

    Examples:
        >>> subtract_numeric = SubtractNumeric()
        >>> subtract_numeric([2, 1, 2], [1, 2, 2]).tolist()
        [1, -1, 0]
    """

    name = "subtract_numeric"
    input_types = [
        ColumnSchema(semantic_tags={"numeric"}),
        ColumnSchema(semantic_tags={"numeric"}),
    ]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    compatibility = [Library.PANDAS, Library.DASK]
    description_template = "the result of {} minus {}"
    commutative = True

    def __init__(self, commutative=True):
        self.commutative = commutative

    def get_function(self):
        return np.subtract

    def generate_name(self, base_feature_names):
        return "%s - %s" % (base_feature_names[0], base_feature_names[1])



from featuretools.primitives.standard.transform.exponential.exponential_weighted_average import (
    ExponentialWeightedAverage,
)
from featuretools.primitives.standard.transform.exponential.exponential_weighted_std import (
    ExponentialWeightedSTD,
)
from featuretools.primitives.standard.transform.exponential.exponential_weighted_variance import (
    ExponentialWeightedVariance,
)



from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import URL, Categorical

from featuretools.primitives.base import TransformPrimitive


class URLToDomain(TransformPrimitive):
    """Determines the domain of a url.

    Description:
        Calculates the label to identify the network domain of a URL. Supports
        urls with or without protocol as well as international country domains.

    Examples:
        >>> url_to_domain = URLToDomain()
        >>> urls =  ['https://play.google.com',
        ...          'http://www.google.co.in',
        ...          'www.facebook.com']
        >>> url_to_domain(urls).tolist()
        ['play.google.com', 'google.co.in', 'facebook.com']
    """

    name = "url_to_domain"
    input_types = [ColumnSchema(logical_type=URL)]
    return_type = ColumnSchema(logical_type=Categorical, semantic_tags={"category"})

    def get_function(self):
        def url_to_domain(x):
            p = r"^(?:https?:\/\/)?(?:[^@\/\n]+@)?(?:www\.)?([^:\/?\n]+)"
            return x.str.extract(p, expand=False)

        return url_to_domain



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import LatLong

from featuretools.primitives.base import TransformPrimitive


class Latitude(TransformPrimitive):
    """Returns the first tuple value in a list of LatLong tuples.
       For use with the LatLong logical type.

    Examples:
        >>> latitude = Latitude()
        >>> latitude([(42.4, -71.1),
        ...            (40.0, -122.4),
        ...            (41.2, -96.75)]).tolist()
        [42.4, 40.0, 41.2]
    """

    name = "latitude"
    input_types = [ColumnSchema(logical_type=LatLong)]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    description_template = "the latitude of {}"

    def get_function(self):
        def latitude(latlong):
            latlong = np.array(latlong.tolist())
            return latlong[:, 0]

        return latitude



import numpy as np
import pandas as pd

from featuretools.primitives import PunctuationCount
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestPunctuationCount(PrimitiveTestBase):
    primitive = PunctuationCount

    def test_punctuation(self):
        x = pd.Series(
            [
                "This is a test file.",
                "This, is second line?",
                "third/line $1,000;",
                "and--subsequen't lines...",
                "*and, more..",
            ],
        )
        primitive_func = self.primitive().get_function()
        answers = [1.0, 2.0, 4.0, 6.0, 4.0]
        np.testing.assert_array_equal(primitive_func(x), answers)

    def test_multiline(self):
        x = pd.Series(
            [
                "This is a test file.",
                "This is second line\nthird line $1000;\nand subsequent lines",
            ],
        )
        primitive_func = self.primitive().get_function()
        answers = [1.0, 2.0]
        np.testing.assert_array_equal(primitive_func(x), answers)

    def test_nan(self):
        x = pd.Series([np.nan, "", "This is a test file."])
        primitive_func = self.primitive().get_function()
        answers = [np.nan, 0.0, 1.0]
        np.testing.assert_array_equal(primitive_func(x), answers)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



import numpy as np
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.utils.gen_utils import Library


class Std(AggregationPrimitive):
    """Computes the dispersion relative to the mean value, ignoring `NaN`.

    Examples:
        >>> std = Std()
        >>> round(std([1, 2, 3, 4, 5, None]), 3)
        1.414
    """

    name = "std"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(semantic_tags={"numeric"})
    stack_on_self = False
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the standard deviation of {}"

    def get_function(self, agg_type=Library.PANDAS):
        if agg_type in [Library.DASK, Library.SPARK]:
            return "std"

        return np.std



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable

from featuretools.primitives.base import AggregationPrimitive


class NumConsecutiveLessMean(AggregationPrimitive):
    """Determines the length of the longest subsequence below the mean.

    Description:
        Given a list of numbers, find the longest subsequence of numbers
        smaller than the mean of the entire sequence. Return the length
        of the longest subsequence.

    Args:
        skipna (bool): If this is False and any value in x is `NaN`, then
            the result will be `NaN`. If True, `NaN` values are skipped.
            Default is True.

    Examples:
        >>> num_consecutive_less_mean = NumConsecutiveLessMean()
        >>> num_consecutive_less_mean([1, 2, 3, 4, 5, 6])
        3.0

        We can also control the way `NaN` values are handled.

        >>> num_consecutive_less_mean = NumConsecutiveLessMean(skipna=False)
        >>> num_consecutive_less_mean([1, 2, 3, 4, 5, 6, None])
        nan
    """

    name = "num_consecutive_less_mean"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def __init__(self, skipna=True):
        self.skipna = skipna

    def get_function(self):
        def num_consecutive_less_mean(x):
            # check for NaN cases
            if x.isnull().all():
                return np.nan
            if not self.skipna and x.isnull().values.any():
                return np.nan
            x_mean = x.mean()

            # In some cases, the mean of x may be NaN
            #   (such as when x has both inf and -inf values)
            if np.isnan(x.mean()):
                return np.nan

            # Find indices of points at or above mean
            x = x.dropna().reset_index(drop=True)
            above_mean_indices = x[x >= x_mean].index.to_series()

            # If none of x is above the mean, return the length of x
            if above_mean_indices.empty:
                return len(x)

            # Pad index with start/end values, in case the longest
            #   sequence occurs at the beginning or end of x
            above_mean_indices[-1] = -1
            above_mean_indices[len(x)] = len(x)
            above_mean_indices = above_mean_indices.sort_index()

            # Calculate gaps between points above mean
            above_mean_indices_shifted = above_mean_indices.shift(1)
            diffs = above_mean_indices - above_mean_indices_shifted

            # Take biggest gap, and subtract 1 to get result
            max_gap = (diffs).max() - 1
            return max_gap

        return num_consecutive_less_mean



from featuretools.primitives import get_aggregation_primitives, get_transform_primitives


def docstring_is_uniform(primitive):
    docstring = primitive.__doc__
    valid_verbs = [
        "Calculates",
        "Determines",
        "Transforms",
        "Computes",
        "Counts",
        "Negates",
        "Adds",
        "Subtracts",
        "Multiplies",
        "Divides",
        "Performs",
        "Returns",
        "Shifts",
        "Extracts",
        "Applies",
    ]
    return any(docstring.startswith(s) for s in valid_verbs)


def test_transform_primitive_docstrings():
    for primitive in get_transform_primitives().values():
        assert docstring_is_uniform(primitive)


def test_aggregation_primitive_docstrings():
    for primitive in get_aggregation_primitives().values():
        assert docstring_is_uniform(primitive)



import pandas as pd
from numpy import random
from numpy.random import choice
from woodwork.logical_types import Categorical, PostalCode

import featuretools as ft


def load_mock_customer(
    n_customers=5,
    n_products=5,
    n_sessions=35,
    n_transactions=500,
    random_seed=0,
    return_single_table=False,
    return_entityset=False,
):
    """Return dataframes of mock customer data"""

    random.seed(random_seed)
    last_date = pd.to_datetime("12/31/2013")
    first_date = pd.to_datetime("1/1/2008")
    first_bday = pd.to_datetime("1/1/1970")

    join_dates = [
        random.uniform(0, 1) * (last_date - first_date) + first_date
        for _ in range(n_customers)
    ]
    birth_dates = [
        random.uniform(0, 1) * (first_date - first_bday) + first_bday
        for _ in range(n_customers)
    ]

    customers_df = pd.DataFrame({"customer_id": range(1, n_customers + 1)})
    customers_df["zip_code"] = choice(
        ["60091", "13244"],
        n_customers,
    )
    customers_df["join_date"] = pd.Series(join_dates).dt.round("1s")
    customers_df["birthday"] = pd.Series(birth_dates).dt.round("1d")

    products_df = pd.DataFrame({"product_id": pd.Categorical(range(1, n_products + 1))})
    products_df["brand"] = choice(["A", "B", "C"], n_products)

    sessions_df = pd.DataFrame({"session_id": range(1, n_sessions + 1)})
    sessions_df["customer_id"] = choice(customers_df["customer_id"], n_sessions)
    sessions_df["device"] = choice(["desktop", "mobile", "tablet"], n_sessions)

    transactions_df = pd.DataFrame({"transaction_id": range(1, n_transactions + 1)})
    transactions_df["session_id"] = choice(sessions_df["session_id"], n_transactions)
    transactions_df = transactions_df.sort_values("session_id").reset_index(drop=True)
    transactions_df["transaction_time"] = pd.date_range(
        "1/1/2014",
        periods=n_transactions,
        freq="65s",
    )  # todo make these less regular
    transactions_df["product_id"] = pd.Categorical(
        choice(products_df["product_id"], n_transactions),
    )
    transactions_df["amount"] = random.randint(500, 15000, n_transactions) / 100

    # calculate and merge in session start
    # based on the times we came up with for transactions
    session_starts = transactions_df.drop_duplicates("session_id")[
        ["session_id", "transaction_time"]
    ].rename(columns={"transaction_time": "session_start"})
    sessions_df = sessions_df.merge(session_starts)

    if return_single_table:
        return (
            transactions_df.merge(sessions_df)
            .merge(customers_df)
            .merge(products_df)
            .reset_index(drop=True)
        )
    elif return_entityset:
        es = ft.EntitySet(id="transactions")
        es = es.add_dataframe(
            dataframe_name="transactions",
            dataframe=transactions_df,
            index="transaction_id",
            time_index="transaction_time",
            logical_types={"product_id": Categorical},
        )

        es = es.add_dataframe(
            dataframe_name="products",
            dataframe=products_df,
            index="product_id",
        )

        es = es.add_dataframe(
            dataframe_name="sessions",
            dataframe=sessions_df,
            index="session_id",
            time_index="session_start",
        )

        es = es.add_dataframe(
            dataframe_name="customers",
            dataframe=customers_df,
            index="customer_id",
            time_index="join_date",
            logical_types={"zip_code": PostalCode},
        )

        rels = [
            ("products", "product_id", "transactions", "product_id"),
            ("sessions", "session_id", "transactions", "session_id"),
            ("customers", "customer_id", "sessions", "customer_id"),
        ]
        es = es.add_relationships(rels)
        es.add_last_time_indexes()
        return es

    return {
        "customers": customers_df,
        "sessions": sessions_df,
        "transactions": transactions_df,
        "products": products_df,
    }



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Boolean, IntegerNullable

from featuretools.primitives.base import AggregationPrimitive


class NumTrueSinceLastFalse(AggregationPrimitive):
    """Calculates the number of 'True' values since the last `False` value.
    Description:
        From a series of Booleans, find the last record with a `False` value.
        Return the count of 'True' values between that record and the end of
        the series. Return nan if no values are `False`. Any nan values in the
        input are ignored. A 'False' value in the last row will result in a
        count of 0.
    Examples:
        >>> num_true_since_last_false = NumTrueSinceLastFalse()
        >>> num_true_since_last_false([False, True, False, True, True])
        2
    """

    name = "num_true_since_last_false"
    input_types = [ColumnSchema(logical_type=Boolean)]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def num_true_since_last_false(x):
            x = x.dropna().astype(bool)
            false_indices = x[~x]
            if false_indices.empty:
                return np.nan
            last_false_index = false_indices.index[-1]
            x_slice = x.loc[last_false_index:]
            return x_slice.sum()

        return num_true_since_last_false



# -*- coding: utf-8 -*-

import re

import numpy as np
import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Double, NaturalLanguage

from featuretools.primitives.base import TransformPrimitive

PUNCTUATION = re.escape("!,.:;?")
END_OF_SENTENCE_PUNCT_RE = re.compile(
    rf"[{PUNCTUATION}]+$|[{PUNCTUATION}]+ |[{PUNCTUATION}]+\n",
)


def _mean_characters_per_word(value):
    if pd.isna(value):
        return np.nan

    # replace end-of-sentence punctuation with space
    value = END_OF_SENTENCE_PUNCT_RE.sub(" ", value)
    words = value.split()
    character_count = [len(x) for x in words]

    return np.mean(character_count) if len(character_count) else 0


class MeanCharactersPerWord(TransformPrimitive):
    """Determines the mean number of characters per word.

    Description:
        Given list of strings, determine the mean number of
        characters per word in each string. A word is defined as
        a series of any characters not separated by white space.
        Punctuation is removed before counting. If a string
        is empty or `NaN`, return `NaN`.

    Examples:
        >>> x = ['This is a test file', 'This is second line', 'third line $1,000']
        >>> mean_characters_per_word = MeanCharactersPerWord()
        >>> mean_characters_per_word(x).tolist()
        [3.0, 4.0, 5.0]
    """

    name = "mean_characters_per_word"
    input_types = [ColumnSchema(logical_type=NaturalLanguage)]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})
    default_value = 0

    def get_function(self):
        def mean_characters_per_word(series):
            return series.apply(_mean_characters_per_word)

        return mean_characters_per_word



import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Categorical, PersonFullName

from featuretools.primitives.base import TransformPrimitive


class FullNameToFirstName(TransformPrimitive):
    """Determines the first name from a person's name.

    Description:
        Given a list of names, determines the first name. If
        only a single name is provided, assume this is a first name.
        If only a title and a single name is provided return `nan`.
        This assumes all titles will be followed by a period. Please note,
        in the current implementation, last names containing spaces may
        result in improper first name matches.


    Examples:
        >>> full_name_to_first_name = FullNameToFirstName()
        >>> names = ['Woolf Spector', 'Oliva y Ocana, Dona. Fermina',
        ...          'Ware, Mr. Frederick', 'Peter, Michael J', 'Mr. Brown']
        >>> full_name_to_first_name(names).to_list()
        ['Woolf', 'Oliva', 'Frederick', 'Michael', nan]
    """

    name = "full_name_to_first_name"
    input_types = [ColumnSchema(logical_type=PersonFullName)]
    return_type = ColumnSchema(logical_type=Categorical, semantic_tags={"category"})

    def get_function(self):
        def full_name_to_first_name(x):
            title_with_last_pattern = r"(^[A-Z][a-z]+\. [A-Z][a-z]+$)"
            titles_pattern = r"([A-Z][a-z]+)\. "
            df = pd.DataFrame({"names": x})
            # remove any entries with just a title and a name
            df["names"] = df["names"].str.replace(
                title_with_last_pattern,
                "",
                regex=True,
            )
            # remove any known titles
            df["names"] = df["names"].str.replace(titles_pattern, "", regex=True)
            # extract first names
            pattern = r"([A-Z][a-z]+ |, [A-Z][a-z]+$|^[A-Z][a-z]+$)"
            df["first_name"] = df["names"].str.extract(pattern)
            # clean up white space and leftover commas
            df["first_name"] = df["first_name"].str.replace(",", "").str.strip()
            return df["first_name"]

        return full_name_to_first_name



import re
import tarfile
from datetime import datetime

import numpy as np
import pandas as pd
from woodwork.logical_types import Datetime, Ordinal

from featuretools.entityset.timedelta import Timedelta


def _check_timedelta(td):
    """
    Convert strings to Timedelta objects
    Allows for both shortform and longform units, as well as any form of capitalization
    '2 Minutes'
    '2 minutes'
    '2 m'
    '1 Minute'
    '1 minute'
    '1 m'
    '1 units'
    '1 Units'
    '1 u'
    Shortform is fine if space is dropped
    '2m'
    '1u"
    If a pd.Timedelta object is passed, units will be converted to seconds due to the underlying representation
        of pd.Timedelta.
    If a pd.DateOffset object is passed, it will be converted to a Featuretools Timedelta if it has one
        temporal parameter. Otherwise, it will remain a pd.DateOffset.
    """
    if td is None:
        return td
    if isinstance(td, Timedelta):
        return td
    elif not isinstance(td, (int, float, str, pd.DateOffset, pd.Timedelta)):
        raise ValueError("Unable to parse timedelta: {}".format(td))
    if isinstance(td, pd.Timedelta):
        unit = "s"
        value = td.total_seconds()
        times = {unit: value}
        return Timedelta(times, delta_obj=td)
    elif isinstance(td, pd.DateOffset):
        # DateOffsets
        if td.__class__.__name__ != "DateOffset":
            if hasattr(td, "__dict__"):
                # Special offsets (such as BDay) - prior to pandas 1.0.0
                value = td.__dict__["n"]
            else:
                # Special offsets (such as BDay) - after pandas 1.0.0
                value = td.n
            unit = td.__class__.__name__
            times = dict([(unit, value)])
        else:
            times = dict()
            for td_unit, td_value in td.kwds.items():
                times[td_unit] = td_value
        return Timedelta(times, delta_obj=td)
    else:
        pattern = "([0-9]+) *([a-zA-Z]+)$"
        match = re.match(pattern, td)
        value, unit = match.groups()
        try:
            value = int(value)
        except Exception:
            try:
                value = float(value)
            except Exception:
                raise ValueError(
                    "Unable to parse value {} from ".format(value)
                    + "timedelta string: {}".format(td),
                )
        times = {unit: value}
        return Timedelta(times)


def _check_time_against_column(time, time_column):
    """
    Check to make sure that time is compatible with time_column,
    where time could be a timestamp, or a Timedelta, number, or None,
    and time_column is a Woodwork initialized column. Compatibility means that
    arithmetic can be performed between time and elements of time_column

    If time is None, then we don't care if arithmetic can be performed
    (presumably it won't ever be performed)
    """
    if time is None:
        return True
    elif isinstance(time, (int, float)):
        return time_column.ww.schema.is_numeric
    elif isinstance(time, (pd.Timestamp, datetime, pd.DateOffset)):
        return time_column.ww.schema.is_datetime
    elif isinstance(time, Timedelta):
        if time_column.ww.schema.is_datetime:
            return True
        elif time.unit not in Timedelta._time_units:
            if (
                isinstance(time_column.ww.logical_type, Ordinal)
                or "numeric" in time_column.ww.semantic_tags
                or "time_index" in time_column.ww.semantic_tags
            ):
                return True
    return False


def _check_time_type(time):
    """
    Checks if `time` is an instance of common int, float, or datetime types.
    Returns "numeric" or Datetime based on results
    """
    time_type = None
    if isinstance(time, (datetime, np.datetime64)):
        time_type = Datetime
    elif (
        isinstance(time, (int, float))
        or np.issubdtype(time, np.integer)
        or np.issubdtype(time, np.floating)
    ):
        time_type = "numeric"
    return time_type


def _is_s3(string):
    """
    Checks if the given string is a s3 path.
    Returns a boolean.
    """
    return string.startswith("s3://")


def _is_url(string):
    """
    Checks if the given string is an url path.
    Returns a boolean.
    """
    return string.startswith("http")


def _is_local_tar(string):
    """
    Checks if the given string is a local tarfile path.
    Returns a boolean.
    """
    return string.endswith(".tar") and tarfile.is_tarfile(string)



import pandas as pd
import pytest

from featuretools.primitives.utils import _get_natural_language_primitives

TIMEOUT_THRESHOLD = 20


class TestNaturalLanguagePrimitivesTerminate:
    # need to sort primitives to avoid pytest collection error
    primitives = sorted(_get_natural_language_primitives().items())

    @pytest.mark.timeout(TIMEOUT_THRESHOLD)
    @pytest.mark.parametrize("primitive", [prim for _, prim in primitives])
    def test_natlang_primitive_does_not_timeout(
        self,
        strings_that_have_triggered_errors_before,
        primitive,
    ):
        for text in strings_that_have_triggered_errors_before:
            primitive().get_function()(pd.Series(text))



import os

import boto3
import pandas as pd
import pytest
from pympler.asizeof import asizeof
from smart_open import open
from woodwork.column_schema import ColumnSchema

from featuretools import (
    AggregationFeature,
    DirectFeature,
    EntitySet,
    Feature,
    GroupByTransformFeature,
    IdentityFeature,
    TransformFeature,
    dfs,
    feature_base,
    load_features,
    primitives,
    save_features,
)
from featuretools.feature_base import FeatureOutputSlice
from featuretools.feature_base.cache import feature_cache
from featuretools.feature_base.features_deserializer import FeaturesDeserializer
from featuretools.feature_base.features_serializer import FeaturesSerializer
from featuretools.primitives import (
    Count,
    CumSum,
    Day,
    DistanceToHoliday,
    Haversine,
    IsIn,
    Max,
    Mean,
    Min,
    Mode,
    Month,
    MultiplyNumericScalar,
    Negate,
    NMostCommon,
    NumberOfCommonWords,
    NumCharacters,
    NumUnique,
    NumWords,
    PercentTrue,
    Skew,
    Std,
    Sum,
    TransformPrimitive,
    Weekday,
    Year,
)
from featuretools.primitives.base import AggregationPrimitive
from featuretools.tests.testing_utils import check_names
from featuretools.utils.gen_utils import Library
from featuretools.version import ENTITYSET_SCHEMA_VERSION, FEATURES_SCHEMA_VERSION

BUCKET_NAME = "test-bucket"
WRITE_KEY_NAME = "test-key"
TEST_S3_URL = "s3://{}/{}".format(BUCKET_NAME, WRITE_KEY_NAME)
TEST_FILE = "test_feature_serialization_feature_schema_{}_entityset_schema_{}_2022_12_28.json".format(
    FEATURES_SCHEMA_VERSION,
    ENTITYSET_SCHEMA_VERSION,
)
S3_URL = "s3://featuretools-static/" + TEST_FILE
URL = "https://featuretools-static.s3.amazonaws.com/" + TEST_FILE
TEST_CONFIG = "CheckConfigPassesOn"
TEST_KEY = "test_access_key_features"


@pytest.fixture(autouse=True)
def reset_dfs_cache():
    feature_cache.enabled = False
    feature_cache.clear_all()


def assert_features(original, deserialized):
    for feat_1, feat_2 in zip(original, deserialized):
        assert feat_1.unique_name() == feat_2.unique_name()
        assert feat_1.entityset == feat_2.entityset

        # IdentityFeature and DirectFeature objects do not have primitives, so
        # series library does not need to be compared
        if not (isinstance(feat_1, (IdentityFeature, DirectFeature))):
            assert feat_1.primitive.series_library == feat_2.primitive.series_library


def pickle_features_test_helper(es_size, features_original, dir_path):
    filepath = os.path.join(dir_path, "test_feature")

    save_features(features_original, filepath)
    features_deserializedA = load_features(filepath)
    assert os.path.getsize(filepath) < es_size
    os.remove(filepath)

    with open(filepath, "w") as f:
        save_features(features_original, f)
    features_deserializedB = load_features(open(filepath))
    assert os.path.getsize(filepath) < es_size
    os.remove(filepath)

    features = save_features(features_original)
    features_deserializedC = load_features(features)
    assert asizeof(features) < es_size

    features_deserialized_options = [
        features_deserializedA,
        features_deserializedB,
        features_deserializedC,
    ]
    for features_deserialized in features_deserialized_options:
        assert_features(features_original, features_deserialized)


def test_pickle_features(es, tmp_path):
    features_original = dfs(
        target_dataframe_name="sessions",
        entityset=es,
        features_only=True,
    )
    pickle_features_test_helper(asizeof(es), features_original, str(tmp_path))


def test_pickle_features_with_custom_primitive(pd_es, tmp_path):
    class NewMax(AggregationPrimitive):
        name = "new_max"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})

    features_original = dfs(
        target_dataframe_name="sessions",
        entityset=pd_es,
        agg_primitives=["Last", "Mean", NewMax],
        features_only=True,
    )

    assert any([isinstance(feat.primitive, NewMax) for feat in features_original])
    pickle_features_test_helper(asizeof(pd_es), features_original, str(tmp_path))


def test_serialized_renamed_features(es):
    def serialize_name_unchanged(original):
        new_name = "MyFeature"
        original_names = original.get_feature_names()
        renamed = original.rename(new_name)
        new_names = (
            [new_name]
            if len(original_names) == 1
            else [new_name + "[{}]".format(i) for i in range(len(original_names))]
        )
        check_names(renamed, new_name, new_names)

        serializer = FeaturesSerializer([renamed])
        serialized = serializer.to_dict()

        deserializer = FeaturesDeserializer(serialized)
        deserialized = deserializer.to_list()[0]
        check_names(deserialized, new_name, new_names)

    identity_original = IdentityFeature(es["log"].ww["value"])
    assert identity_original.get_name() == "value"

    value = IdentityFeature(es["log"].ww["value"])

    primitive = primitives.Max()
    agg_original = AggregationFeature(value, "customers", primitive)
    assert agg_original.get_name() == "MAX(log.value)"

    direct_original = DirectFeature(
        IdentityFeature(es["customers"].ww["age"]),
        "sessions",
    )
    assert direct_original.get_name() == "customers.age"

    primitive = primitives.MultiplyNumericScalar(value=2)
    transform_original = TransformFeature(value, primitive)
    assert transform_original.get_name() == "value * 2"

    zipcode = IdentityFeature(es["log"].ww["zipcode"])
    primitive = CumSum()
    groupby_original = feature_base.GroupByTransformFeature(value, primitive, zipcode)
    assert groupby_original.get_name() == "CUM_SUM(value) by zipcode"

    multioutput_original = Feature(
        es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )
    assert multioutput_original.get_name() == "N_MOST_COMMON(log.product_id, n=2)"

    featureslice_original = feature_base.FeatureOutputSlice(multioutput_original, 0)
    assert featureslice_original.get_name() == "N_MOST_COMMON(log.product_id, n=2)[0]"

    feature_type_list = [
        identity_original,
        agg_original,
        direct_original,
        transform_original,
        groupby_original,
        multioutput_original,
        featureslice_original,
    ]

    for feature_type in feature_type_list:
        serialize_name_unchanged(feature_type)


@pytest.fixture
def s3_client():
    _environ = os.environ.copy()
    from moto import mock_s3

    with mock_s3():
        s3 = boto3.resource("s3")
        yield s3
    os.environ.clear()
    os.environ.update(_environ)


@pytest.fixture
def s3_bucket(s3_client, region="us-east-2"):
    location = {"LocationConstraint": region}
    s3_client.create_bucket(
        Bucket=BUCKET_NAME,
        ACL="public-read-write",
        CreateBucketConfiguration=location,
    )
    s3_bucket = s3_client.Bucket(BUCKET_NAME)
    yield s3_bucket


def test_serialize_features_mock_s3(es, s3_client, s3_bucket):
    features_original = dfs(
        target_dataframe_name="sessions",
        entityset=es,
        features_only=True,
    )

    save_features(features_original, TEST_S3_URL)

    obj = list(s3_bucket.objects.all())[0].key
    s3_client.ObjectAcl(BUCKET_NAME, obj).put(ACL="public-read-write")

    features_deserialized = load_features(TEST_S3_URL)
    assert_features(features_original, features_deserialized)


def test_serialize_features_mock_anon_s3(es, s3_client, s3_bucket):
    features_original = dfs(
        target_dataframe_name="sessions",
        entityset=es,
        features_only=True,
    )

    save_features(features_original, TEST_S3_URL, profile_name=False)

    obj = list(s3_bucket.objects.all())[0].key
    s3_client.ObjectAcl(BUCKET_NAME, obj).put(ACL="public-read-write")

    features_deserialized = load_features(TEST_S3_URL, profile_name=False)
    assert_features(features_original, features_deserialized)


@pytest.mark.parametrize("profile_name", ["test", False])
def test_s3_test_profile(es, s3_client, s3_bucket, setup_test_profile, profile_name):
    features_original = dfs(
        target_dataframe_name="sessions",
        entityset=es,
        features_only=True,
    )

    save_features(features_original, TEST_S3_URL, profile_name="test")

    obj = list(s3_bucket.objects.all())[0].key
    s3_client.ObjectAcl(BUCKET_NAME, obj).put(ACL="public-read-write")

    features_deserialized = load_features(TEST_S3_URL, profile_name=profile_name)
    assert_features(features_original, features_deserialized)


@pytest.mark.parametrize("url,profile_name", [(S3_URL, False), (URL, None)])
def test_deserialize_features_s3(pd_es, url, profile_name):
    agg_primitives = [
        Sum,
        Std,
        Max,
        Skew,
        Min,
        Mean,
        Count,
        PercentTrue,
        NumUnique,
        Mode,
    ]

    trans_primitives = [Day, Year, Month, Weekday, Haversine, NumWords, NumCharacters]

    features_original = dfs(
        target_dataframe_name="sessions",
        entityset=pd_es,
        features_only=True,
        agg_primitives=agg_primitives,
        trans_primitives=trans_primitives,
    )

    features_deserialized = load_features(url, profile_name=profile_name)
    assert_features(features_original, features_deserialized)


def test_serialize_url(es):
    features_original = dfs(
        target_dataframe_name="sessions",
        entityset=es,
        features_only=True,
    )
    error_text = "Writing to URLs is not supported"
    with pytest.raises(ValueError, match=error_text):
        save_features(features_original, URL)


def test_custom_feature_names_retained_during_serialization(pd_es, tmp_path):
    class MultiCumulative(TransformPrimitive):
        name = "multi_cum_sum"
        input_types = [ColumnSchema(semantic_tags={"numeric"})]
        return_type = ColumnSchema(semantic_tags={"numeric"})
        number_output_features = 3

    multi_output_trans_feat = Feature(
        pd_es["log"].ww["value"],
        primitive=MultiCumulative,
    )
    groupby_trans_feat = GroupByTransformFeature(
        pd_es["log"].ww["value"],
        primitive=MultiCumulative,
        groupby=pd_es["log"].ww["product_id"],
    )
    multi_output_agg_feat = Feature(
        pd_es["log"].ww["product_id"],
        parent_dataframe_name="customers",
        primitive=NMostCommon(n=2),
    )
    slice = FeatureOutputSlice(multi_output_trans_feat, 1)
    stacked_feat = Feature(slice, primitive=Negate)

    trans_names = ["cumulative_sum", "cumulative_max", "cumulative_min"]
    multi_output_trans_feat.set_feature_names(trans_names)
    groupby_trans_names = ["grouped_sum", "grouped_max", "grouped_min"]
    groupby_trans_feat.set_feature_names(groupby_trans_names)
    agg_names = ["first_most_common", "second_most_common"]
    multi_output_agg_feat.set_feature_names(agg_names)

    features = [
        multi_output_trans_feat,
        multi_output_agg_feat,
        groupby_trans_feat,
        stacked_feat,
    ]
    file = os.path.join(tmp_path, "features.json")
    save_features(features, file)
    deserialized_features = load_features(file)

    new_trans, new_agg, new_groupby, new_stacked = deserialized_features
    assert new_trans.get_feature_names() == trans_names
    assert new_agg.get_feature_names() == agg_names
    assert new_groupby.get_feature_names() == groupby_trans_names
    assert new_stacked.get_feature_names() == ["-(cumulative_max)"]


def test_deserializer_uses_common_primitive_instances_no_args(es, tmp_path):
    features = dfs(
        entityset=es,
        target_dataframe_name="products",
        features_only=True,
        agg_primitives=["sum"],
        trans_primitives=["is_null"],
    )

    is_null_features = [f for f in features if f.primitive.name == "is_null"]
    sum_features = [f for f in features if f.primitive.name == "sum"]

    # Make sure we have multiple features of each type
    assert len(is_null_features) > 1
    assert len(sum_features) > 1

    # DFS should use the same primitive instance for all features that share a primitive
    is_null_primitive = is_null_features[0].primitive
    sum_primitive = sum_features[0].primitive
    assert all([f.primitive is is_null_primitive for f in is_null_features])
    assert all([f.primitive is sum_primitive for f in sum_features])

    file = os.path.join(tmp_path, "features.json")
    save_features(features, file)
    deserialized_features = load_features(file)
    new_is_null_features = [
        f for f in deserialized_features if f.primitive.name == "is_null"
    ]
    new_sum_features = [f for f in deserialized_features if f.primitive.name == "sum"]

    # After deserialization all features that share a primitive should use the same primitive instance
    new_is_null_primitive = new_is_null_features[0].primitive
    new_sum_primitive = new_sum_features[0].primitive
    assert all([f.primitive is new_is_null_primitive for f in new_is_null_features])
    assert all([f.primitive is new_sum_primitive for f in new_sum_features])


def test_deserializer_uses_common_primitive_instances_with_args(es, tmp_path):
    # Single argument
    scalar1 = MultiplyNumericScalar(value=1)
    scalar5 = MultiplyNumericScalar(value=5)
    features = dfs(
        entityset=es,
        target_dataframe_name="products",
        features_only=True,
        agg_primitives=["sum"],
        trans_primitives=[scalar1, scalar5],
    )

    scalar1_features = [
        f
        for f in features
        if f.primitive.name == "multiply_numeric_scalar" and " * 1" in f.get_name()
    ]
    scalar5_features = [
        f
        for f in features
        if f.primitive.name == "multiply_numeric_scalar" and " * 5" in f.get_name()
    ]

    # Make sure we have multiple features of each type
    assert len(scalar1_features) > 1
    assert len(scalar5_features) > 1

    # DFS should use the the passed in primitive instance for all features
    assert all([f.primitive is scalar1 for f in scalar1_features])
    assert all([f.primitive is scalar5 for f in scalar5_features])

    file = os.path.join(tmp_path, "features.json")
    save_features(features, file)
    deserialized_features = load_features(file)

    new_scalar1_features = [
        f
        for f in deserialized_features
        if f.primitive.name == "multiply_numeric_scalar" and " * 1" in f.get_name()
    ]
    new_scalar5_features = [
        f
        for f in deserialized_features
        if f.primitive.name == "multiply_numeric_scalar" and " * 5" in f.get_name()
    ]

    # After deserialization all features that share a primitive should use the same primitive instance
    new_scalar1_primitive = new_scalar1_features[0].primitive
    new_scalar5_primitive = new_scalar5_features[0].primitive
    assert all([f.primitive is new_scalar1_primitive for f in new_scalar1_features])
    assert all([f.primitive is new_scalar5_primitive for f in new_scalar5_features])
    assert new_scalar1_primitive.value == 1
    assert new_scalar5_primitive.value == 5

    # Test primitive with multiple args - pandas only due to primitive compatibility
    if es.dataframe_type == Library.PANDAS:
        distance_to_holiday = DistanceToHoliday(
            holiday="Victoria Day",
            country="Canada",
        )
        features = dfs(
            entityset=es,
            target_dataframe_name="customers",
            features_only=True,
            agg_primitives=[],
            trans_primitives=[distance_to_holiday],
        )

        distance_features = [
            f for f in features if f.primitive.name == "distance_to_holiday"
        ]

        assert len(distance_features) > 1

        # DFS should use the the passed in primitive instance for all features
        assert all([f.primitive is distance_to_holiday for f in distance_features])

        file = os.path.join(tmp_path, "distance_features.json")
        save_features(distance_features, file)
        new_distance_features = load_features(file)

        # After deserialization all features that share a primitive should use the same primitive instance
        new_distance_primitive = new_distance_features[0].primitive
        assert all(
            [f.primitive is new_distance_primitive for f in new_distance_features],
        )
        assert new_distance_primitive.holiday == "Victoria Day"
        assert new_distance_primitive.country == "Canada"

    # Test primitive with list arg
    is_in = IsIn(list_of_outputs=[5, True, "coke zero"])
    features = dfs(
        entityset=es,
        target_dataframe_name="customers",
        features_only=True,
        agg_primitives=[],
        trans_primitives=[is_in],
    )

    is_in_features = [f for f in features if f.primitive.name == "isin"]
    assert len(is_in_features) > 1

    # DFS should use the the passed in primitive instance for all features
    assert all([f.primitive is is_in for f in is_in_features])

    file = os.path.join(tmp_path, "distance_features.json")
    save_features(is_in_features, file)
    new_is_in_features = load_features(file)

    # After deserialization all features that share a primitive should use the same primitive instance
    new_is_in_primitive = new_is_in_features[0].primitive
    assert all([f.primitive is new_is_in_primitive for f in new_is_in_features])
    assert new_is_in_primitive.list_of_outputs == [5, True, "coke zero"]


def test_can_serialize_word_set_for_number_of_common_words_feature(pd_es):
    # The word_set argument is passed in as a set, which is not JSON-serializable.
    # This test checks internal logic that converts the set to a list so it can be serialized
    common_word_set = {"hello", "my"}
    df = pd.DataFrame({"text": ["hello my name is hi"]})
    es = EntitySet()
    es.add_dataframe(dataframe_name="df", index="idx", dataframe=df, make_index=True)

    num_common_words = NumberOfCommonWords(word_set=common_word_set)
    fm, fd = dfs(
        entityset=es,
        target_dataframe_name="df",
        trans_primitives=[num_common_words],
    )

    feat = fd[-1]
    save_features([feat])



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive


class CountBelowMean(AggregationPrimitive):
    """Determines the number of values that are below the mean.

    Args:
        skipna (bool): Determines if to use NA/null values. Defaults to
            True to skip NA/null.

    Examples:
        >>> count_below_mean = CountBelowMean()
        >>> count_below_mean([1, 2, 3, 4, 10])
        3

        The way NaNs are treated can be controlled.

        >>> count_below_mean_skipna = CountBelowMean(skipna=False)
        >>> count_below_mean_skipna([1, 2, 3, 4, 5, None])
        nan
    """

    name = "count_below_mean"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    stack_on_self = False

    def __init__(self, skipna=True):
        self.skipna = skipna

    def get_function(self):
        def count_below_mean(x):
            mean = x.mean(skipna=self.skipna)
            if np.isnan(mean):
                return np.nan
            return len(x[x < mean])

        return count_below_mean



from inspect import signature

import pytest

from featuretools import (
    FeatureBase,
    calculate_feature_matrix,
    dfs,
    encode_features,
    list_primitives,
    load_features,
    save_features,
)
from featuretools.primitives.base import AggregationPrimitive, PrimitiveBase
from featuretools.tests.testing_utils import make_ecommerce_entityset

PRIMITIVES = list_primitives()


def get_number_from_offset(offset):
    """Extract the numeric element of a potential offset string.

    Args:
        offset (int, str): If offset is an integer, that value is returned. If offset is a string,
            it's assumed to be an offset string of the format nD where n is a single digit integer.

    Note: This helper utility should only be used with offset strings that only have one numeric character.
        Only the first character will be returned, so if an offset string 24H is used, it will incorrectly
        return the integer 2. Additionally, any of the offset timespans (H for hourly, D for daily, etc.)
        can be used here; however, care should be taken by the user to remember what that timespan is when
        writing tests, as comparing 7 from 7D to 1 from 1W may not behave as expected.
    """
    if isinstance(offset, str):
        return int(offset[0])
    else:
        return offset


class PrimitiveTestBase:
    primitive = None

    @pytest.fixture()
    def es(self):
        es = make_ecommerce_entityset()
        return es

    def test_name_and_desc(self):
        assert self.primitive.name is not None
        assert self.primitive.__doc__ is not None
        docstring = self.primitive.__doc__
        short_description = docstring.splitlines()[0]
        first_word = short_description.split(" ", 1)[0]
        valid_verbs = [
            "Calculates",
            "Determines",
            "Transforms",
            "Computes",
            "Shifts",
            "Extracts",
            "Applies",
        ]
        assert any(s in first_word for s in valid_verbs)
        assert self.primitive.input_types is not None

    def test_name_in_primitive_list(self):
        assert PRIMITIVES.name.eq(self.primitive.name).any()

    def test_arg_init(self):
        primitive_ = self.primitive()
        # determine the optional arguments in the __init__
        init_params = signature(self.primitive.__init__)
        for name, parameter in init_params.parameters.items():
            if parameter.default is not parameter.empty:
                assert hasattr(primitive_, name)

    def test_serialize(self, es, target_dataframe_name="log"):
        check_serialize(primitive=self.primitive, es=es, target_dataframe_name="log")


def check_serialize(primitive, es, target_dataframe_name="log"):
    trans_primitives = []
    agg_primitives = []
    if issubclass(primitive, AggregationPrimitive):
        agg_primitives = [primitive]
    else:
        trans_primitives = [primitive]
    features = dfs(
        entityset=es,
        target_dataframe_name=target_dataframe_name,
        agg_primitives=agg_primitives,
        trans_primitives=trans_primitives,
        max_features=-1,
        max_depth=3,
        features_only=True,
        return_types="all",
    )

    feat_to_serialize = None
    for feature in features:
        if feature.primitive.__class__ == primitive:
            feat_to_serialize = feature
            break
        for base_feature in feature.get_dependencies(deep=True):
            if base_feature.primitive.__class__ == primitive:
                feat_to_serialize = base_feature
                break
    assert feat_to_serialize is not None

    # Skip calculating feature matrix for long running primitives
    skip_primitives = ["elmo"]

    if primitive.name not in skip_primitives:
        df1 = calculate_feature_matrix([feat_to_serialize], entityset=es)

    new_feat = load_features(save_features([feat_to_serialize]))[0]
    assert isinstance(new_feat, FeatureBase)

    if primitive.name not in skip_primitives:
        df2 = calculate_feature_matrix([new_feat], entityset=es)
        assert df1.equals(df2)


def find_applicable_primitives(primitive):
    from featuretools.primitives.utils import (
        get_aggregation_primitives,
        get_transform_primitives,
    )

    all_transform_primitives = list(get_transform_primitives().values())
    all_aggregation_primitives = list(get_aggregation_primitives().values())
    applicable_transforms = find_stackable_primitives(
        all_transform_primitives,
        primitive,
    )
    applicable_aggregations = find_stackable_primitives(
        all_aggregation_primitives,
        primitive,
    )
    return applicable_transforms, applicable_aggregations


def find_stackable_primitives(all_primitives, primitive):
    applicable_primitives = []
    for x in all_primitives:
        if x.input_types == [primitive.return_type]:
            applicable_primitives.append(x)
    return applicable_primitives


def valid_dfs(
    es,
    aggregations,
    transforms,
    feature_substrings,
    target_dataframe_name="log",
    multi_output=False,
    max_depth=3,
    max_features=-1,
    instance_ids=[0, 1, 2, 3],
):
    if not isinstance(feature_substrings, list):
        feature_substrings = [feature_substrings]

    if any([issubclass(x, PrimitiveBase) for x in feature_substrings]):
        feature_substrings = [x.name.upper() for x in feature_substrings]

    features = dfs(
        entityset=es,
        target_dataframe_name=target_dataframe_name,
        agg_primitives=aggregations,
        trans_primitives=transforms,
        max_features=max_features,
        max_depth=max_depth,
        features_only=True,
    )
    applicable_features = []
    for feat in features:
        applicable_features += [
            feat for x in feature_substrings if x in feat.get_name()
        ]
    if len(applicable_features) == 0:
        raise ValueError(
            "No feature names with %s, verify the name attribute \
                          is defined and/or generate_name() is defined to \
                          return %s "
            % (feature_substrings, feature_substrings),
        )
    df = calculate_feature_matrix(
        entityset=es,
        features=applicable_features,
        instance_ids=instance_ids,
        n_jobs=1,
    )

    encode_features(df, applicable_features)

    # TODO: check the multi_output shape by checking
    # feature.number_output_features for each feature
    # and comparing it with the matrix shape
    if not multi_output:
        assert len(applicable_features) == df.shape[1]
    return



"""
dfs_profile.py

Helper module to allow profiling of the dfs operations.  At some point we may
want to use pstats to output the results to a log, but I'm anticipating that
LookingGlass will provide the performance data we want.

Notes:
  - output currently goes to the root directory and is in dfs_profile.stats
  - *.stats is gitignored
  - it uses the demo customers dataset for testing
  - max_depth > 2 is very slow (currently)
  - stats output can be viewed online with https://nejc.saje.info/pstats-viewer.html
"""
import cProfile
from pathlib import Path

import featuretools as ft
import featuretools.demo as demo
from featuretools.synthesis.dfs import dfs

es = demo.load_retail()

all_aggs = ft.primitives.get_aggregation_primitives()
all_trans = ft.primitives.get_transform_primitives()

profiler = cProfile.Profile(builtins=False)
profiler.enable()
feature_defs = dfs(
    entityset=es,
    target_dataframe_name="customers",
    trans_primitives=all_trans,
    agg_primitives=all_aggs,
    max_depth=2,
    features_only=True,
)
profiler.disable()
profiler.dump_stats(Path.cwd() / "dfs_profile.stats")



import numpy as np
import pandas as pd

from featuretools.primitives import CountString
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestCountString(PrimitiveTestBase):
    primitive = CountString

    def compare(self, primitive_initiated, test_cases, answers):
        primitive_func = primitive_initiated.get_function()
        primitive_answers = primitive_func(test_cases)
        return np.testing.assert_array_equal(answers, primitive_answers)

    test_cases = pd.Series(
        [
            # Ignore case
            "Hello other words hello hEllo HELLO",
            # ignore non alphanumeric
            "he\\{ll\t\n\t.--?o othe/r words hello hello h.el./lo",
            # match whole word
            "hellohellohello other hello word go hello here 9hello hello9",
            # all combined
            #   hello/ counts as hello being it's own word
            #   since * and / are non word characters
            #   but 9 is a "word character" so 9hello9
            #   does not count as hello being its own word
            "helloHellohello 9Hello 9hello9 *hello/ test'hel..lo' 'hE.l.lO' \
         hello",
        ],
    )

    def test_non_regex_with_no_other_parameters(self):
        primitive = self.primitive(
            "hello",
            ignore_case=False,
            ignore_non_alphanumeric=False,
            is_regex=False,
            match_whole_words_only=False,
        )
        answers = [1, 2, 7, 5]
        self.compare(primitive, self.test_cases, answers)

    def test_non_regex_ignore_case(self):
        primitive1 = self.primitive(
            "hello",
            ignore_case=True,
            ignore_non_alphanumeric=False,
            is_regex=False,
            match_whole_words_only=False,
        )

        primitive2 = self.primitive(
            "HeLLo",
            ignore_case=True,
            ignore_non_alphanumeric=False,
            is_regex=False,
            match_whole_words_only=False,
        )

        answers = [4, 2, 7, 7]
        self.compare(primitive1, self.test_cases, answers)
        self.compare(primitive2, self.test_cases, answers)

    def test_non_regex_ignore_non_alphanumeric(self):
        primitive = self.primitive(
            "hello",
            ignore_case=False,
            ignore_non_alphanumeric=True,
            is_regex=False,
            match_whole_words_only=False,
        )
        answers = [1, 4, 7, 6]
        self.compare(primitive, self.test_cases, answers)

    def test_non_regex_match_whole_words_only(self):
        primitive = self.primitive(
            "hello",
            ignore_case=False,
            ignore_non_alphanumeric=False,
            is_regex=False,
            match_whole_words_only=True,
        )

        answers = [1, 2, 2, 2]
        self.compare(primitive, self.test_cases, answers)

    def test_non_regex_with_all_others_parameters(self):
        primitive = self.primitive(
            "hello",
            ignore_case=True,
            ignore_non_alphanumeric=True,
            is_regex=False,
            match_whole_words_only=True,
        )

        answers = [4, 4, 2, 3]
        self.compare(primitive, self.test_cases, answers)

    def test_regex_with_no_other_parameters(self):
        primitive = self.primitive(
            "h.l.o",
            ignore_case=False,
            ignore_non_alphanumeric=False,
            is_regex=True,
            match_whole_words_only=False,
        )

        answers = [2, 2, 7, 5]
        self.compare(primitive, self.test_cases, answers)

    def test_regex_with_ignore_case(self):
        primitive = self.primitive(
            "h.l.o",
            ignore_case=True,
            ignore_non_alphanumeric=False,
            is_regex=True,
            match_whole_words_only=False,
        )

        answers = [4, 2, 7, 7]
        self.compare(primitive, self.test_cases, answers)

    def test_regex_with_ignore_non_alphanumeric(self):
        primitive = self.primitive(
            "h.l.o",
            ignore_case=False,
            ignore_non_alphanumeric=True,
            is_regex=True,
            match_whole_words_only=False,
        )

        answers = [2, 4, 7, 6]
        self.compare(primitive, self.test_cases, answers)

    def test_regex_with_match_whole_words_only(self):
        primitive = self.primitive(
            "h.l.o",
            ignore_case=False,
            ignore_non_alphanumeric=False,
            is_regex=True,
            match_whole_words_only=True,
        )

        answers = [2, 2, 2, 2]
        self.compare(primitive, self.test_cases, answers)

    def test_regex_with_all_other_parameters(self):
        primitive = self.primitive(
            "h.l.o",
            ignore_case=True,
            ignore_non_alphanumeric=True,
            is_regex=True,
            match_whole_words_only=True,
        )

        answers = [4, 4, 2, 3]
        self.compare(primitive, self.test_cases, answers)

    def test_overlapping_regex(self):
        primitive = self.primitive(
            "(?=(a.*a))",
            ignore_case=True,
            ignore_non_alphanumeric=True,
            is_regex=True,
            match_whole_words_only=False,
        )
        test_cases = pd.Series(["aaaaaaaaaa", "atesta aa aa a"])
        answers = [9, 6]
        self.compare(primitive, test_cases, answers)

    def test_the(self):
        primitive = self.primitive(
            "the",
            ignore_case=True,
            ignore_non_alphanumeric=False,
            is_regex=False,
            match_whole_words_only=False,
        )
        test_cases = pd.Series(["The fox jumped over the cat", "The there then"])

        answers = [2, 3]
        self.compare(primitive, test_cases, answers)

    def test_nan(self):
        primitive = self.primitive(
            "the",
            ignore_case=True,
            ignore_non_alphanumeric=False,
            is_regex=False,
            match_whole_words_only=False,
        )
        test_cases = pd.Series(
            [np.nan, None, pd.NA, "The fox jumped over the cat", "The there then"],
        )
        answers = [np.nan, np.nan, np.nan, 2, 3]
        self.compare(primitive, test_cases, answers)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive(
            "the",
            ignore_case=True,
            ignore_non_alphanumeric=False,
            is_regex=False,
            match_whole_words_only=False,
        )
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)

    def test_with_featuretools_nan(self, es):
        log_df = es["log"]
        comments = log_df["comments"]
        comments[1] = pd.NA
        comments[2] = np.nan
        comments[3] = None
        log_df["comments"] = comments
        es.replace_dataframe(dataframe_name="log", df=log_df)

        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive(
            "the",
            ignore_case=True,
            ignore_non_alphanumeric=False,
            is_regex=False,
            match_whole_words_only=False,
        )
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



import numpy as np
import pandas as pd

from featuretools.primitives import UpperCaseCount
from featuretools.tests.primitive_tests.utils import (
    PrimitiveTestBase,
    find_applicable_primitives,
    valid_dfs,
)


class TestUpperCaseCount(PrimitiveTestBase):
    primitive = UpperCaseCount

    def test_strings(self):
        x = pd.Series(
            ["This IS a STRING.", "Testing AaA", "Testing AAA-BBB", "testing aaa"],
        )
        primitive_func = self.primitive().get_function()
        answers = [9.0, 3.0, 7.0, 0.0]
        np.testing.assert_array_equal(primitive_func(x), answers)

    def test_nan(self):
        x = pd.Series([np.nan, "", "This IS a STRING."])
        primitive_func = self.primitive().get_function()
        answers = [np.nan, 0.0, 9.0]
        np.testing.assert_array_equal(primitive_func(x), answers)

    def test_with_featuretools(self, es):
        transform, aggregation = find_applicable_primitives(self.primitive)
        primitive_instance = self.primitive()
        transform.append(primitive_instance)
        valid_dfs(es, aggregation, transform, self.primitive)



import pandas as pd
import pytest

from featuretools import dfs


@pytest.fixture
def pd_entry_point_dfs():
    cards_df = pd.DataFrame({"id": [1, 2, 3, 4, 5]})
    transactions_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5, 6],
            "card_id": [1, 2, 1, 3, 4, 5],
            "transaction_time": [10, 12, 13, 20, 21, 20],
            "fraud": [True, False, True, False, True, True],
        },
    )
    return cards_df, transactions_df


@pytest.fixture
def dask_entry_point_dfs(pd_entry_point_dfs):
    dd = pytest.importorskip("dask.dataframe", reason="Dask not installed, skipping")
    cards_df = dd.from_pandas(pd_entry_point_dfs[0], npartitions=2)
    transactions_df = dd.from_pandas(pd_entry_point_dfs[1], npartitions=2)
    return cards_df, transactions_df


@pytest.fixture(params=["pd_entry_point_dfs", "dask_entry_point_dfs"])
def entry_points_dfs(request):
    return request.getfixturevalue(request.param)


class MockEntryPoint(object):
    def on_call(self, kwargs):
        self.kwargs = kwargs

    def on_error(self, error, runtime):
        self.error = error

    def on_return(self, return_value, runtime):
        self.return_value = return_value

    def load(self):
        return self

    def __call__(self):
        return self


class MockPkgResources(object):
    def __init__(self, entry_point):
        self.entry_point = entry_point

    def iter_entry_points(self, name):
        return [self.entry_point]


def test_entry_point(es, monkeypatch):
    entry_point = MockEntryPoint()
    # overrides a module used in the entry_point decorator for dfs
    # so the decorator will use this mock entry point
    monkeypatch.setitem(
        dfs.__globals__["entry_point"].__globals__,
        "pkg_resources",
        MockPkgResources(entry_point),
    )
    fm, fl = dfs(entityset=es, target_dataframe_name="customers")
    assert "entityset" in entry_point.kwargs.keys()
    assert "target_dataframe_name" in entry_point.kwargs.keys()
    assert (fm, fl) == entry_point.return_value


def test_entry_point_error(es, monkeypatch):
    entry_point = MockEntryPoint()
    monkeypatch.setitem(
        dfs.__globals__["entry_point"].__globals__,
        "pkg_resources",
        MockPkgResources(entry_point),
    )
    with pytest.raises(KeyError):
        dfs(entityset=es, target_dataframe_name="missing_dataframe")

    assert isinstance(entry_point.error, KeyError)


def test_entry_point_detect_arg(monkeypatch, entry_points_dfs):
    cards_df = entry_points_dfs[0]
    transactions_df = entry_points_dfs[1]
    cards_df = pd.DataFrame({"id": [1, 2, 3, 4, 5]})
    transactions_df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5, 6],
            "card_id": [1, 2, 1, 3, 4, 5],
            "transaction_time": [10, 12, 13, 20, 21, 20],
            "fraud": [True, False, True, False, True, True],
        },
    )
    dataframes = {
        "cards": (cards_df, "id"),
        "transactions": (transactions_df, "id", "transaction_time"),
    }
    relationships = [("cards", "id", "transactions", "card_id")]
    entry_point = MockEntryPoint()
    monkeypatch.setitem(
        dfs.__globals__["entry_point"].__globals__,
        "pkg_resources",
        MockPkgResources(entry_point),
    )
    fm, fl = dfs(dataframes, relationships, target_dataframe_name="cards")
    assert "dataframes" in entry_point.kwargs.keys()
    assert "relationships" in entry_point.kwargs.keys()
    assert "target_dataframe_name" in entry_point.kwargs.keys()



import pandas as pd
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.utils.gen_utils import Library, import_or_none

dd = import_or_none("dask.dataframe")


class NumUnique(AggregationPrimitive):
    """Determines the number of distinct values, ignoring `NaN` values.

    Examples:
        >>> num_unique = NumUnique()
        >>> num_unique(['red', 'blue', 'green', 'yellow'])
        4

        `NaN` values will be ignored.

        >>> num_unique(['red', 'blue', 'green', 'yellow', None])
        4
    """

    name = "num_unique"
    input_types = [ColumnSchema(semantic_tags={"category"})]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    stack_on_self = False
    compatibility = [Library.PANDAS, Library.DASK, Library.SPARK]
    description_template = "the number of unique elements in {}"

    def get_function(self, agg_type=Library.PANDAS):
        if agg_type == Library.DASK:

            def chunk(s):
                def inner_chunk(x):
                    x = x[:].dropna()
                    return set(x.unique())

                return s.agg(inner_chunk)

            def agg(s):
                def inner_agg(x):
                    x = x[:].dropna()
                    return set().union(*x.values)

                return s.agg(inner_agg)

            def finalize(s):
                return s.apply(lambda x: len(x))

            return dd.Aggregation(self.name, chunk=chunk, agg=agg, finalize=finalize)

        elif agg_type == Library.SPARK:
            return "nunique"

        return pd.Series.nunique



import numpy as np
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import IntegerNullable

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive


class CountOutsideRange(AggregationPrimitive):
    """Determines the number of values that fall outside a certain range.

    Args:
        lower (float): Lower boundary of range (exclusive). Default is 0.
        upper (float): Upper boundary of range (exclusive). Default is 1.
        skipna (bool): Determines if to use NA/null values. Defaults to
            True to skip NA/null.

    Examples:
        >>> count_outside_range = CountOutsideRange(lower=1.5, upper=3.6)
        >>> count_outside_range([1, 2, 3, 4, 5])
        3

        The way NaNs are treated can be controlled.

        >>> count_outside_range_skipna = CountOutsideRange(skipna=False)
        >>> count_outside_range_skipna([1, 2, 3, 4, 5, None])
        nan
    """

    name = "count_outside_range"
    input_types = [ColumnSchema(semantic_tags={"numeric"})]
    return_type = ColumnSchema(logical_type=IntegerNullable, semantic_tags={"numeric"})
    stack_on_self = False
    default_value = 0

    def __init__(self, lower=0, upper=1, skipna=True):
        self.lower = lower
        self.upper = upper
        self.skipna = skipna

    def get_function(self):
        def count_outside_range(x):
            if not self.skipna and x.isnull().values.any():
                return np.nan
            cond = (x < self.lower) | (x > self.upper)
            return cond.sum()

        return count_outside_range



import json

import featuretools as ft


def describe_feature(
    feature,
    feature_descriptions=None,
    primitive_templates=None,
    metadata_file=None,
):
    """Generates an English language description of a feature.

    Args:
        feature (FeatureBase) : Feature to describe
        feature_descriptions (dict, optional) : dictionary mapping features or unique
            feature names to custom descriptions
        primitive_templates (dict, optional) : dictionary mapping primitives or
            primitive names to description templates
        metadata_file (str, optional) : path to json metadata file

    Returns:
        str : English description of the feature
    """
    feature_descriptions = feature_descriptions or {}
    primitive_templates = primitive_templates or {}

    if metadata_file:
        file_feature_descriptions, file_primitive_templates = parse_json_metadata(
            metadata_file,
        )
        feature_descriptions = {**file_feature_descriptions, **feature_descriptions}
        primitive_templates = {**file_primitive_templates, **primitive_templates}

    description = generate_description(
        feature,
        feature_descriptions,
        primitive_templates,
    )
    return description[:1].upper() + description[1:] + "."


def generate_description(feature, feature_descriptions, primitive_templates):
    # Check if feature has custom description
    if feature in feature_descriptions or feature.unique_name() in feature_descriptions:
        description = feature_descriptions.get(feature) or feature_descriptions.get(
            feature.unique_name(),
        )
        return description

    # Check if identity feature:
    if isinstance(feature, ft.IdentityFeature):
        description = feature.column_schema.description
        if description is None:
            description = 'the "{}"'.format(feature.column_name)
        return description

    # Handle direct features
    if isinstance(feature, ft.DirectFeature):
        base_feature, direct_description = get_direct_description(feature)
        direct_base = generate_description(
            base_feature,
            feature_descriptions,
            primitive_templates,
        )
        return direct_base + direct_description

    # Get input descriptions
    input_descriptions = []
    input_columns = feature.base_features
    if isinstance(feature, ft.feature_base.FeatureOutputSlice):
        input_columns = feature.base_feature.base_features

    for input_col in input_columns:
        col_description = generate_description(
            input_col,
            feature_descriptions,
            primitive_templates,
        )
        input_descriptions.append(col_description)

    # Remove groupby description from input columns
    groupby_description = None
    if isinstance(feature, ft.GroupByTransformFeature):
        groupby_description = input_descriptions.pop()

    # Generate primitive description
    template_override = None
    if (
        feature.primitive in primitive_templates
        or feature.primitive.name in primitive_templates
    ):
        template_override = primitive_templates.get(
            feature.primitive,
        ) or primitive_templates.get(feature.primitive.name)
    slice_num = feature.n if hasattr(feature, "n") else None
    primitive_description = feature.primitive.get_description(
        input_descriptions,
        slice_num=slice_num,
        template_override=template_override,
    )
    if isinstance(feature, ft.feature_base.FeatureOutputSlice):
        feature = feature.base_feature

    # Generate groupby phrase if applicable
    groupby = ""
    if isinstance(feature, ft.AggregationFeature):
        groupby_description = get_aggregation_groupby(feature, feature_descriptions)
    if groupby_description is not None:
        if groupby_description.startswith("the "):
            groupby_description = groupby_description[4:]
        groupby = "for each {}".format(groupby_description)

    # Generate aggregation dataframe phrase with use_previous
    dataframe_description = ""
    if isinstance(feature, ft.AggregationFeature):
        if feature.use_previous:
            dataframe_description = "of the previous {} of ".format(
                feature.use_previous.get_name().lower(),
            )
        else:
            dataframe_description = "of all instances of "
        dataframe_description += '"{}"'.format(
            feature.relationship_path[-1][1].child_dataframe.ww.name,
        )

    # Generate where phrase
    where = ""
    if hasattr(feature, "where") and feature.where:
        where_col = generate_description(
            feature.where.base_features[0],
            feature_descriptions,
            primitive_templates,
        )
        where = "where {} is {}".format(where_col, feature.where.primitive.value)

    # Join all parts of template
    description_template = [
        primitive_description,
        dataframe_description,
        where,
        groupby,
    ]
    description = " ".join([phrase for phrase in description_template if phrase != ""])

    return description


def get_direct_description(feature):
    direct_description = (
        ' the instance of "{}" associated with this '
        'instance of "{}"'.format(
            feature.relationship_path[-1][1].parent_dataframe.ww.name,
            feature.dataframe_name,
        )
    )
    base_features = feature.base_features
    # shortens stacked direct features to make it easier to understand
    while isinstance(base_features[0], ft.DirectFeature):
        base_feat = base_features[0]
        base_feat_description = ' the instance of "{}" associated ' "with".format(
            base_feat.relationship_path[-1][1].parent_dataframe.ww.name,
        )
        direct_description = base_feat_description + direct_description
        base_features = base_feat.base_features
    direct_description = " for" + direct_description

    return base_features[0], direct_description


def get_aggregation_groupby(feature, feature_descriptions=None):
    if feature_descriptions is None:
        feature_descriptions = {}
    groupby_name = feature.dataframe.ww.index
    groupby = ft.IdentityFeature(
        feature.entityset[feature.dataframe_name].ww[groupby_name],
    )
    if groupby in feature_descriptions or groupby.unique_name() in feature_descriptions:
        return feature_descriptions.get(groupby) or feature_descriptions.get(
            groupby.unique_name(),
        )
    else:
        return '"{}" in "{}"'.format(groupby_name, feature.dataframe_name)


def parse_json_metadata(file):
    with open(file) as f:
        json_metadata = json.load(f)

    return (
        json_metadata.get("feature_descriptions", {}),
        json_metadata.get("primitive_templates", {}),
    )



from featuretools.feature_base import (
    AggregationFeature,
    FeatureOutputSlice,
    GroupByTransformFeature,
    TransformFeature,
)
from featuretools.utils.gen_utils import camel_and_title_to_snake


def _categorize_features(features):
    """Categorize each feature by its primitive type in a set of primitives along with any dependencies"""
    transform = set()
    agg = set()
    groupby = set()
    where = set()
    explored = set()

    def get_feature_data(feature):
        if feature.get_name() in explored:
            return

        dependencies = []

        if isinstance(feature, FeatureOutputSlice):
            feature = feature.base_feature

        if isinstance(feature, AggregationFeature):
            if feature.where:
                where.add(feature.primitive.name)
            else:
                agg.add(feature.primitive.name)
        elif isinstance(feature, GroupByTransformFeature):
            groupby.add(feature.primitive.name)
        elif isinstance(feature, TransformFeature):
            transform.add(feature.primitive.name)

        feature_deps = feature.get_dependencies()
        if feature_deps:
            dependencies.extend(feature_deps)

        explored.add(feature.get_name())

        for dep in dependencies:
            get_feature_data(dep)

    for feature in features:
        get_feature_data(feature)

    return transform, agg, groupby, where


def get_unused_primitives(specified, used):
    """Get a list of unused primitives based on a list of specified primitives and a list of output features"""
    if not specified:
        return []
    specified = {
        camel_and_title_to_snake(primitive)
        if isinstance(primitive, str)
        else primitive.name
        for primitive in specified
    }
    return sorted(specified.difference(used))



from datetime import datetime

import numpy as np
import pandas as pd
import pytest

from featuretools.primitives import DateToHoliday


def test_datetoholiday():
    date_to_holiday = DateToHoliday()

    dates = pd.Series(
        [
            datetime(2016, 1, 1),
            datetime(2016, 2, 27),
            datetime(2017, 5, 29, 10, 30, 5),
            datetime(2018, 7, 4),
        ],
    )

    holiday_series = date_to_holiday(dates).tolist()

    assert holiday_series[0] == "New Year's Day"
    assert np.isnan(holiday_series[1])
    assert holiday_series[2] == "Memorial Day"
    assert holiday_series[3] == "Independence Day"


def test_datetoholiday_error():
    error_text = r"must be one of the available countries.*"
    with pytest.raises(ValueError, match=error_text):
        DateToHoliday(country="UNK")


def test_nat():
    date_to_holiday = DateToHoliday()
    case = pd.Series(
        [
            "2019-10-14",
            "NaT",
            "2016-02-15",
            "NaT",
        ],
    ).astype("datetime64[ns]")
    answer = ["Columbus Day", np.nan, "Washington's Birthday", np.nan]
    given_answer = date_to_holiday(case).astype("str")
    np.testing.assert_array_equal(given_answer, answer)


def test_valid_country():
    date_to_holiday = DateToHoliday(country="Canada")
    case = pd.Series(
        [
            "2016-07-01",
            "2016-11-11",
            "2017-12-26",
            "2018-09-03",
        ],
    ).astype("datetime64[ns]")
    answer = ["Canada Day", np.nan, "Boxing Day", "Labour Day"]
    given_answer = date_to_holiday(case).astype("str")
    np.testing.assert_array_equal(given_answer, answer)


def test_multiple_countries():
    dth_mexico = DateToHoliday(country="Mexico")

    case = pd.Series([datetime(2000, 9, 16), datetime(2005, 1, 1)])
    assert len(dth_mexico(case)) > 1

    dth_india = DateToHoliday(country="IND")
    case = pd.Series([datetime(2048, 1, 1), datetime(2048, 10, 2)])
    assert len(dth_india(case)) > 1

    dth_uk = DateToHoliday(country="UK")
    case = pd.Series([datetime(2048, 3, 17), datetime(2048, 4, 6)])
    assert len(dth_uk(case)) > 1

    countries = [
        "Argentina",
        "AU",
        "Austria",
        "BY",
        "Belgium",
        "Brazil",
        "Canada",
        "Colombia",
        "Croatia",
        "England",
        "Finland",
        "FRA",
        "Germany",
        "Germany",
        "Italy",
        "NewZealand",
        "PortugalExt",
        "PTE",
        "Spain",
        "ES",
        "Switzerland",
        "UnitedStates",
        "US",
        "UK",
        "UA",
        "CH",
        "SE",
        "ZA",
    ]
    for x in countries:
        DateToHoliday(country=x)


def test_with_timezone_aware_datetimes():
    df = pd.DataFrame(
        {
            "non_timezone_aware_with_time": pd.date_range(
                "2018-07-03 09:00",
                periods=3,
            ),
            "non_timezone_aware_no_time": pd.date_range("2018-07-03", periods=3),
            "timezone_aware_with_time": pd.date_range(
                "2018-07-03 09:00",
                periods=3,
            ).tz_localize(tz="US/Eastern"),
            "timezone_aware_no_time": pd.date_range(
                "2018-07-03",
                periods=3,
            ).tz_localize(tz="US/Eastern"),
        },
    )

    date_to_holiday = DateToHoliday(country="US")
    expected = [np.nan, "Independence Day", np.nan]
    for col in df.columns:
        actual = date_to_holiday(df[col]).astype("str")
        np.testing.assert_array_equal(actual, expected)



from pandas import NaT
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime

from featuretools.primitives.base import AggregationPrimitive


class DateFirstEvent(AggregationPrimitive):
    """Determines the first datetime from a list of datetimes.

    Examples:
        >>> from datetime import datetime
        >>> date_first_event = DateFirstEvent()
        >>> date_first_event([
        ...     datetime(2011, 4, 9, 10, 30, 10),
        ...     datetime(2011, 4, 9, 10, 30, 20),
        ...     datetime(2011, 4, 9, 10, 30, 30)])
        Timestamp('2011-04-09 10:30:10')
    """

    name = "date_first_event"
    input_types = [ColumnSchema(logical_type=Datetime, semantic_tags={"time_index"})]
    return_type = ColumnSchema(logical_type=Datetime)
    stack_on_self = False
    default_value = 0

    def get_function(self):
        def date_first_event(x):
            x = x.dropna()
            if x.empty:
                return NaT
            return x.iat[0]

        return date_first_event



import numpy as np
import pandas as pd
from woodwork import init_series
from woodwork.column_schema import ColumnSchema
from woodwork.logical_types import Datetime, Double

from featuretools.primitives.base.transform_primitive_base import TransformPrimitive
from featuretools.primitives.standard.transform.time_series.utils import (
    apply_rolling_agg_to_series,
)


class RollingOutlierCount(TransformPrimitive):
    """Determines how many values are outliers over a given window.

    Description:
        Given a list of numbers and a corresponding list of
        datetimes, return a rolling count of outliers within the numeric values,
        starting at the row `gap` rows away from the current row and looking backward
        over the specified window (by `window_length` and `gap`). Values are deemed
        outliers using the IQR method, computed over the whole series.
        Input datetimes should be monotonic.

    Args:
        window_length (int, string, optional): Specifies the amount of data included in each window.
            If an integer is provided, it will correspond to a number of rows. For data with a uniform sampling
            frequency, for example of one day, the window_length will correspond to a period of time, in this case,
            7 days for a window_length of 7.
            If a string is provided, it must be one of Pandas' offset alias strings ('1D', '1H', etc),
            and it will indicate a length of time that each window should span.
            The list of available offset aliases can be found at
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases.
            Defaults to 3.
        gap (int, string, optional): Specifies a gap backwards from each instance before the
            window of usable data begins. If an integer is provided, it will correspond to a number of rows.
            If a string is provided, it must be one of Pandas' offset alias strings ('1D', '1H', etc),
            and it will indicate a length of time between a target instance and the beginning of its window.
            Defaults to 1, which excludes the target instance from the window.
        min_periods (int, optional): Minimum number of observations required for performing calculations
            over the window. Can only be as large as window_length when window_length is an integer.
            When window_length is an offset alias string, this limitation does not exist, but care should be taken
            to not choose a min_periods that will always be larger than the number of observations in a window.
            Defaults to 1.

    Note:
        Only offset aliases with fixed frequencies can be used when defining gap and window_length.
        This means that aliases such as `M` or `W` cannot be used, as they can indicate different
        numbers of days. ('M', because different months are different numbers of days;
        'W' because week will indicate a certain day of the week, like W-Wed, so that will
        indicate a different number of days depending on the anchoring date.)

    Note:
        When using an offset alias to define `gap`, an offset alias must also be used to define `window_length`.
        This limitation does not exist when using an offset alias to define `window_length`. In fact,
        if the data has a uniform sampling frequency, it is preferable to use a numeric `gap` as it is more
        efficient.

    Examples:
        >>> import pandas as pd
        >>> rolling_outlier_count = RollingOutlierCount(window_length=4)
        >>> times = pd.date_range(start='2019-01-01', freq='1min', periods=6)
        >>> rolling_outlier_count(times, [0, 0, 0, 0, 10, 0]).tolist()
        [nan, 0.0, 0.0, 0.0, 0.0, 1.0]

        We can also control the gap before the rolling calculation.
        >>> import pandas as pd
        >>> rolling_outlier_count = RollingOutlierCount(window_length=4, gap=0)
        >>> times = pd.date_range(start='2019-01-01', freq='1min', periods=6)
        >>> rolling_outlier_count(times, [0, 0, 0, 0, 10, 0]).tolist()
        [0.0, 0.0, 0.0, 0.0, 1.0, 1.0]

        We can also control the minimum number of periods required for the rolling calculation.
        >>> import pandas as pd
        >>> rolling_outlier_count = RollingOutlierCount(window_length=4, min_periods=3)
        >>> times = pd.date_range(start='2019-01-01', freq='1min', periods=6)
        >>> rolling_outlier_count(times,  [0, 0, 0, 0, 10, 0]).tolist()
        [nan, nan, nan, 0.0, 0.0, 1.0]

        We can also set the window_length and gap using offset alias strings.
        >>> import pandas as pd
        >>> rolling_outlier_count = RollingOutlierCount(window_length='4min', gap='1min')
        >>> times = pd.date_range(start='2019-01-01', freq='1min', periods=6)
        >>> rolling_outlier_count(times, [0, 0, 0, 0, 10, 0]).tolist()
        [nan, 0.0, 0.0, 0.0, 0.0, 1.0]
    """

    name = "rolling_outlier_count"
    input_types = [
        ColumnSchema(logical_type=Datetime, semantic_tags={"time_index"}),
        ColumnSchema(semantic_tags={"numeric"}),
    ]
    return_type = ColumnSchema(logical_type=Double, semantic_tags={"numeric"})
    uses_full_dataframe = True

    def __init__(self, window_length=3, gap=1, min_periods=0):
        self.window_length = window_length
        self.gap = gap
        self.min_periods = min_periods

    def get_outliers_count(self, numeric_series):
        # We know the column is numeric, so use the Double logical type in case Woodwork's
        # type inference could not infer a numeric type
        if not len(numeric_series.dropna()):
            return np.nan
        if numeric_series.ww.schema is None:
            numeric_series = init_series(numeric_series, logical_type="Double")
        box_plot_info = numeric_series.ww.box_plot_dict()
        return len(box_plot_info["high_values"]) + len(box_plot_info["low_values"])

    def get_function(self):
        def rolling_outlier_count(datetime, numeric):
            x = pd.Series(numeric.values, index=datetime.values)
            return apply_rolling_agg_to_series(
                series=x,
                agg_func=self.get_outliers_count,
                window_length=self.window_length,
                gap=self.gap,
                min_periods=self.min_periods,
                ignore_window_nans=False,
            )

        return rolling_outlier_count



import numpy as np
from woodwork.column_schema import ColumnSchema

from featuretools.primitives.base.aggregation_primitive_base import AggregationPrimitive
from featuretools.utils.gen_utils import Library


class NMostCommon(AggregationPrimitive):
    """Determines the `n` most common elements.

    Description:
        Given a list of values, return the `n` values
        which appear the most frequently. If there are
        fewer than `n` unique values, the output will be
        filled with `NaN`.

    Args:
        n (int): defines "n" in "n most common." Defaults
            to 3.

    Examples:
        >>> n_most_common = NMostCommon(n=2)
        >>> x = ['orange', 'apple', 'orange', 'apple', 'orange', 'grapefruit']
        >>> n_most_common(x).tolist()
        ['orange', 'apple']
    """

    name = "n_most_common"
    input_types = [ColumnSchema(semantic_tags={"category"})]
    return_type = None

    def __init__(self, n=3):
        self.n = n
        self.number_output_features = n
        self.description_template = [
            "the {} most common values of {{}}".format(n),
            "the most common value of {}",
            *["the {nth_slice} most common value of {}"] * (n - 1),
        ]

    def get_function(self, agg_type=Library.PANDAS):
        def n_most_common(x):
            # Counts of 0 remain in value_counts output if dtype is category
            # so we need to remove them
            counts = x.value_counts()
            counts = counts[counts > 0]
            array = np.array(counts.index[: self.n])
            if len(array) < self.n:
                filler = np.full(self.n - len(array), np.nan)
                array = np.append(array, filler)
            return array

        return n_most_common



import pandas as pd
from woodwork.logical_types import BooleanNullable

import featuretools as ft


def test_percent_true_default_value_with_dfs():
    es = ft.EntitySet(id="customer_data")

    customers_df = pd.DataFrame(data={"customer_id": [1, 2]})
    transactions_df = pd.DataFrame(
        data={"tx_id": [1], "customer_id": [1], "is_foo": [True]},
    )

    es.add_dataframe(
        dataframe_name="customers_df",
        dataframe=customers_df,
        index="customer_id",
    )
    es.add_dataframe(
        dataframe_name="transactions_df",
        dataframe=transactions_df,
        index="tx_id",
        logical_types={"is_foo": BooleanNullable},
    )

    es = es.add_relationship(
        "customers_df",
        "customer_id",
        "transactions_df",
        "customer_id",
    )

    feature_matrix, _ = ft.dfs(
        entityset=es,
        target_dataframe_name="customers_df",
        agg_primitives=["percent_true"],
    )

    assert pd.isna(feature_matrix["PERCENT_TRUE(transactions_df.is_foo)"][2])
